"""String utilities."""

# ruff: noqa: RUF001
from __future__ import annotations

from functools import cache
from itertools import takewhile
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TextIO, cast, overload
import os
import re
import string

from bs4 import BeautifulSoup, Tag
import soupsieve as sv

from .typing import StrPath, assert_not_none

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence

__all__ = (
    'add_unidecode_custom_replacement',
    'cssq',
    'cssq_one',
    'fix_apostrophes',
    'fullwidth_to_narrow',
    'hexstr2bytes',
    'hexstr2bytes_generator',
    'is_ascii',
    'is_roman_numeral',
    'is_url',
    'rev_sentences',
    'sanitize',
    'slugify',
    'strip_ansi',
    'strip_ansi_if_no_colors',
    'underscorize',
    'unix_path_to_wine',
)

ORD_MAX = 128
STRIP_ANSI_PATTERN = re.compile(r'\x1B\[\d+(;\d+){0,2}m')


@cache
def strip_ansi(o: str) -> str:
    """
    Remove ANSI escape sequences from ``o``.

    As defined by ECMA-048 in http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-048.

    Taken from https://github.com/ewen-lbh/python-strip-ansi/ due to installation issues with
    Poetry.

    Parameters
    ----------
    o : str
        String to strip ANSI escape sequences from.

    Returns
    -------
    str
        The string without ANSI escape sequences.
    """
    return STRIP_ANSI_PATTERN.sub('', o)


def strip_ansi_if_no_colors(s: str) -> str:
    """
    Strip ANSI colour-codes if the ``NO_COLOR`` environment variable is set.

    See https://no-color.org/.

    Parameters
    ----------
    s : str
        String to conditionally strip.

    Returns
    -------
    str
        The string without ANSI escape sequences if ``NO_COLOR`` is set, otherwise unchanged.
    """
    return strip_ansi(s) if os.environ.get('NO_COLOR') else s


@cache
def underscorize(s: str) -> str:
    """
    Replace all space-type characters with ``_``.

    Parameters
    ----------
    s : str
        String to transform.

    Returns
    -------
    str
        The transformed string.
    """
    return re.sub(r'\s+', '_', s)


@cache
def is_ascii(s: Sequence[str]) -> bool:
    """
    Check if a string consists of only ASCII characters.

    Parameters
    ----------
    s : Sequence[str]
        String to check.

    Returns
    -------
    bool
        ``True`` if all characters are ASCII.
    """
    return len(s) == len(list(takewhile(lambda x: ord(x) < ORD_MAX, s)))


def hexstr2bytes_generator(s: str) -> Iterator[int]:
    """
    Convert a hex string such as ``"01020a"`` to integers.

    Parameters
    ----------
    s : str
        Hex string to convert.

    Yields
    ------
    int
        The integers in the hex string.

    Raises
    ------
    ValueError
        If the string is not a valid hex string.
    """
    for i in range(0, len(s), 2):
        hex_num = s[i:i + 2]
        if len(hex_num) != 2:  # noqa: PLR2004
            raise ValueError(hex_num)
        yield int(hex_num, 16)


def hexstr2bytes(s: str) -> bytes:
    """
    Convert a hex string such as ``"01020a"`` to its bytes form (``0x1 0x2 0x10)``).

    Parameters
    ----------
    s : str
        Hex string to convert.

    Returns
    -------
    bytes
        The bytes representation.
    """
    return bytes(hexstr2bytes_generator(s))


@cache
def unix_path_to_wine(path: StrPath) -> str:
    """
    Convert a UNIX path to an absolute Wine path.

    If the path does not exist, the output will be the current path and the path passed in combined.

    The output path will begin at letter ``Z:``. Other drive letters are not supported.

    Parameters
    ----------
    path : StrPath
        Path to convert.

    Returns
    -------
    str
        Window-style Wine absolute path.
    """
    try:
        path = Path(path).resolve(strict=True)
    except FileNotFoundError:
        path = Path.cwd() / path
    return f'Z:{path}'.replace('/', '\\')


def _get_yt_dlp_sanitize_filename() -> Callable[..., str]:  # pragma: no cover
    from yt_dlp.utils import sanitize_filename  # noqa: PLC0415

    return sanitize_filename


@cache
def sanitize(s: str, *, restricted: bool = True) -> str:
    """
    Transform a string to a 'sanitised' form.

    Parameters
    ----------
    s : str
        String to transform.
    restricted : bool
        If ``True``, use a restricted form. This is suitable for filenames on Windows.

    Returns
    -------
    str
        Returns a transformed string, which will be at minimum ``'_'``.
    """
    sanitize_filename = _get_yt_dlp_sanitize_filename()
    return re.sub(
        r'([a-z0-9])\-s\-',
        r'\1s-',
        re.sub(r'\.-', '-',
               re.sub(r'[_\-]+', '-',
                      sanitize_filename(s, restricted=restricted).lower())),
    )


@cache
def is_url(filename: StrPath) -> bool:
    """
    Detect if ``filename`` is a URL.

    This is the same method mpv uses to decide this.

    Parameters
    ----------
    filename : StrPath
        Path or string to check.

    Returns
    -------
    bool
        ``True`` if ``filename`` looks like a URL.
    """
    parts = str(filename).split('://', 1)
    if len(parts) < 2:  # noqa: PLR2004
        return False
    # protocol prefix has no special characters => it's a URL
    return all(x in f'{string.ascii_letters}{string.digits}_' for x in parts[0])


def _get_unidecode_cache_and_unidecode(
) -> tuple[dict[int, Sequence[str | None] | None], Callable[..., str]]:  # pragma: no cover
    from unidecode import Cache, unidecode  # noqa: PLC0415

    return Cache, unidecode


def add_unidecode_custom_replacement(find: str, replace: str) -> None:
    """
    Add a custom replacement to the Unidecode cache.

    Parameters
    ----------
    find : str
        Single character to find.
    replace : str
        Replacement string.
    """
    cache, unidecode = _get_unidecode_cache_and_unidecode()
    unidecode(find)  # Force it to load the module
    codepoint = ord(find)
    section = codepoint >> 8
    position = codepoint % 256
    new_section = cast(
        'list[str | None]',
        (cache[section] if isinstance(cache[section], list) else
         (list(assert_not_none(cache[section]))
          if cache[section] is not None else [None for _ in range(position + 1)])),
    )  # convert to mutable type
    assert len(new_section) > position
    new_section[position] = replace
    cache[section] = new_section


FULLWIDTH_MAP = (
    ('　', ' '),
    ('…', '...'),
    ('！', '!'),
    ('？', '?'),
    ('～', '~'),
    ('（', '('),
    ('）', ')'),
    ('＂', '"'),
    ('＃', '#'),
    ('＄', '$'),
    ('％', '%'),
    ('＆', '&'),
    ('＇', "'"),
    ('＊', '*'),
    ('，', ','),
    ('－', '-'),
    ('．', '.'),
    ('／', '/'),
    ('０', '0'),
    ('１', '1'),
    ('２', '2'),
    ('３', '3'),
    ('４', '4'),
    ('５', '5'),
    ('６', '6'),
    ('７', '7'),
    ('８', '8'),
    ('９', '9'),
    ('：', ':'),
    ('；', ';'),
    ('＜', '<'),
    ('＝', '='),
    ('＞', '>'),
    ('？', '?'),
    ('＠', '@'),
    ('Ａ', 'A'),
    ('Ｂ', 'B'),
    ('Ｃ', 'C'),
    ('Ｄ', 'D'),
    ('Ｅ', 'E'),
    ('Ｆ', 'F'),
    ('Ｇ', 'G'),
    ('Ｈ', 'H'),
    ('Ｉ', 'I'),
    ('Ｊ', 'J'),
    ('Ｋ', 'K'),
    ('Ｌ', 'L'),
    ('Ｍ', 'M'),
    ('Ｎ', 'N'),
    ('Ｏ', 'O'),
    ('Ｐ', 'P'),
    ('Ｑ', 'Q'),
    ('Ｒ', 'R'),
    ('Ｓ', 'S'),
    ('Ｔ', 'T'),
    ('Ｕ', 'U'),
    ('Ｖ', 'V'),
    ('Ｗ', 'W'),
    ('Ｘ', 'X'),
    ('Ｙ', 'Y'),
    ('Ｚ', 'Z'),
    ('［', '['),
    ('＼', '\\'),
    ('］', ']'),
    ('＾', '^'),
    ('＿', '_'),
    ('｀', '`'),
    ('ａ', 'a'),
    ('ｂ', 'b'),
    ('ｃ', 'c'),
    ('ｄ', 'd'),
    ('ｅ', 'e'),
    ('ｆ', 'f'),
    ('ｇ', 'g'),
    ('ｈ', 'h'),
    ('ｉ', 'i'),
    ('ｊ', 'j'),
    ('ｋ', 'k'),
    ('ｌ', 'l'),
    ('ｍ', 'm'),
    ('ｎ', 'n'),
    ('ｏ', 'o'),
    ('ｐ', 'p'),
    ('ｑ', 'q'),
    ('ｒ', 'r'),
    ('ｓ', 's'),
    ('ｔ', 't'),
    ('ｕ', 'u'),
    ('ｖ', 'v'),
    ('ｗ', 'w'),
    ('ｘ', 'x'),
    ('ｙ', 'y'),
    ('ｚ', 'z'),
    ('｛', '{'),
    ('｜', '|'),
    ('｝', '}'),
    ('｟', '⸨'),
    ('｠', '⸩'),
    ('￠', '¢'),
    ('￡', '£'),
    ('￢', '¬'),
    ('￣', '‾'),
    ('￤', '|'),
    ('￥', '¥'),
    ('￦', '₩'),
)


@cache
def fullwidth_to_narrow(s: str) -> str:
    """
    Convert fullwidth characters in ``s`` to narrow or halfwidth.

    Unlike Unidecode this will convert ``'￥'`` to its halfwidth form ``'¥'``.

    Parameters
    ----------
    s : str
        String containing fullwidth characters.

    Returns
    -------
    str
        The string with fullwidth characters replaced by their narrow equivalents.
    """
    for find, replace in FULLWIDTH_MAP:
        s = s.replace(find, replace)
    return s


@cache
def slugify(s: str, *, no_lower: bool = False) -> str:
    """
    Generate a slug string.

    Parameters
    ----------
    s : str
        String to slugify.
    no_lower : bool
        If ``True``, do not convert to lowercase.

    Returns
    -------
    str
        The slugified string.
    """
    return re.sub(r'[-\s_]+', '-', re.sub(r'[^\w\s-]', '', s if no_lower else s.lower()).strip())


@cache
def is_roman_numeral(string: str) -> bool:
    """
    Check if a string is a valid Roman numeral.

    Parameters
    ----------
    string : str
        String to check.

    Returns
    -------
    bool
        ``True`` if the string is a valid Roman numeral.
    """
    if not string:
        return False
    return (re.match(r'^M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})$', string,
                     re.IGNORECASE) is not None)


@cache
def fix_apostrophes(word: str) -> str:
    """
    Title-case a word while preserving apostrophe contractions.

    Parameters
    ----------
    word : str
        Word to fix.

    Returns
    -------
    str
        The word with correct casing, or unchanged if it contains no apostrophe.
    """
    if "'" not in word:
        return word
    return re.sub(
        r"[A-Za-z]+('[A-Za-z]+)?",
        lambda mo: f'{mo.group(0)[0].upper()}{mo.group(0)[1:].lower()}',
        word,
    )


def rev_sentence(w: str) -> str:
    ending = m.group(1) if (m := re.search(r'([\.\!\?])$', w)) else '.'
    lst = list(reversed(re.sub(r'([\.\!\?:;])$', '', w).split()))
    lst[0] = lst[0].title()
    lst[-1] = lst[-1].lower()
    fixed_i = re.sub(r'(\b)i(\b)', r'\1I\2', ' '.join(lst))
    return f'{fixed_i}{ending}'


def rev_sentences(sentences: Sequence[str]) -> Iterator[str]:
    """
    Reverse the word order of each sentence.

    Parameters
    ----------
    sentences : Sequence[str]
        Sentences to reverse.

    Yields
    ------
    str
        Each sentence with its words in reversed order.
    """
    for line in (x.strip() for x in sentences if x.strip()):
        yield rev_sentence(line)


@overload
def cssq_one(
    selector: str,
    file: TextIO | str,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: Literal[False] = False,
) -> Tag | None:  # pragma: no cover
    ...


@overload
def cssq_one(
    selector: str,
    file: TextIO | str,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: Literal[True] = True,
) -> str | None:  # pragma: no cover
    ...


@overload
def cssq_one(
    selector: str,
    file: TextIO | str,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: bool = False,
) -> str | Tag | None:  # pragma: no cover
    ...


def cssq_one(
    selector: str,
    file: TextIO | str,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: bool = False,
) -> str | Tag | None:
    """
    Select a single item from HTML with CSS.

    Parameters
    ----------
    selector : str
        CSS selector.
    file : TextIO | str
        File-like object or string containing HTML.
    debug_selector : bool
        Enable SoupSieve debug logging.
    strip : bool
        Strip whitespace from text content.
    text : bool
        Output content within the HTML tags only.

    Returns
    -------
    str | Tag | None
        The selected item, or ``None`` if no match was found.
    """
    soup = BeautifulSoup(file, 'html5lib')
    element = soup.select_one(selector, flags=sv.DEBUG if debug_selector else 0)
    if element and text:
        return element.get_text(strip=strip)
    return element


@overload
def cssq(
    selector: str,
    file: TextIO | str,
    limit: int = 0,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: Literal[False] = False,
) -> Iterator[Tag]:  # pragma: no cover
    ...


@overload
def cssq(
    selector: str,
    file: TextIO | str,
    limit: int = 0,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: Literal[True] = True,
) -> Iterator[str]:  # pragma: no cover
    ...


@overload
def cssq(
    selector: str,
    file: TextIO | str,
    limit: int = 0,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: bool = False,
) -> Iterator[str] | Iterator[Tag]:  # pragma: no cover
    ...


def cssq(
    selector: str,
    file: TextIO | str,
    limit: int = 0,
    *,
    debug_selector: bool = False,
    strip: bool = True,
    text: bool = False,
) -> Iterator[Tag] | Iterator[str]:
    """
    Filter HTML with CSS.

    If passing ``limit=1``, consider using :py:func:`cssq_one` instead.

    Parameters
    ----------
    selector : str
        CSS selector.
    file : TextIO | str
        File-like object or string containing HTML.
    limit : int
        Limit number of results. ``0`` means no limit.
    debug_selector : bool
        Enable SoupSieve debug logging.
    strip : bool
        Strip whitespace from text content.
    text : bool
        Output content within the HTML tags only.

    Yields
    ------
    Iterator[Tag] | Iterator[str]
        The filtered items.
    """
    soup = BeautifulSoup(file, 'html5lib')
    yield from [
        item if not text else item.get_text(strip=strip)
        for item in sv.iselect(selector, soup, flags=sv.DEBUG if debug_selector else 0, limit=limit)
    ]



"""Chromium-related functions."""

from __future__ import annotations

from functools import cache
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, cast

import requests

if TYPE_CHECKING:
    from types import ModuleType

    from .typing import StrPath

__all__ = (
    'fix_chromium_pwa_icon',
    'generate_chrome_user_agent',
    'get_last_chrome_major_version',
    'get_latest_chrome_major_version',
)


def _get_pil_image_module() -> ModuleType:  # pragma: no cover
    from PIL import Image  # noqa: PLC0415

    return Image


def fix_chromium_pwa_icon(
    config_path: StrPath,
    app_id: str,
    icon_src_uri: str,
    profile: str = 'Default',
    *,
    masked: bool = False,
    monochrome: bool = False,
) -> None:
    """
    Fix a Chromium PWA icon that failed to sync.

    Parameters
    ----------
    config_path : StrPath
        Path to the Chromium configuration directory.
    app_id : str
        App ID of the PWA.
    icon_src_uri : str
        URI of the icon source.
    profile : str
        Profile name. Default is ``'Default'``.
    masked : bool
        If ``True``, save the icon as a maskable icon. Default is ``False``.
    monochrome : bool
        If ``True``, save the icon as a monochrome icon. Default is ``False``.

    Raises
    ------
    ValueError
        If the icon is not square.

    See Also
    --------
    `Bug 40595456 - PWA icons can be lost (on sync?) and reverted to a letter`_
    """
    image_mod = _get_pil_image_module()
    config_path = Path(config_path) / profile / 'Web Applications' / app_id
    r = requests.get(icon_src_uri, timeout=15)
    r.raise_for_status()
    img = image_mod.open(BytesIO(r.content))
    width, height = img.size
    if width != height:
        msg = 'Icon is not square.'
        raise ValueError(msg)
    sizes = list(reversed([1 << x for x in range(4, min(10, width.bit_length()))]))
    for size in sizes:
        img.resize((size, size), image_mod.LANCZOS).save(config_path / 'Icons' / f'{size}.png')
    if masked:
        for size in sizes:
            img.resize((size, size),
                       image_mod.LANCZOS).save(config_path / 'Icons Maskable' / f'{size}.png')
    if monochrome:
        for size in sizes:
            img.resize((size, size),
                       image_mod.LANCZOS).save(config_path / 'Icons Monochrome' / f'{size}.png')


@cache
def get_last_chrome_major_version() -> str:
    """
    Get last major Chrome version from the profile directory.

    Tries the following, in order:

    * Chrome Beta
    * Chrome
    * Chrome Canary
    * Chromium

    Returns
    -------
    str
        If no ``Last Version`` file is found, returns empty string.
    """
    for location in (
            '~/.config/google-chrome-beta',
            '~/AppData/Local/Google/Chrome Beta/User Data',
            '~/Library/Application Support/Google/Chrome Beta',
            '~/.config/google-chrome',
            '~/AppData/Local/Google/Chrome/User Data',
            '~/Library/Application Support/Google/Chrome',
            '~/.config/google-chrome-unstable',
            '~/AppData/Local/Google/Chrome Canary/User Data',
            '~/Library/Application Support/Google/Chrome Canary',
            '~/.config/chromium',
            '~/AppData/Local/Google/Chromium/User Data',
            '~/Library/Application Support/Google/Chromium',
    ):
        if (p := (Path(location).expanduser() / 'Last Version')).exists():
            return p.read_text().split('.', 1)[0]
    return ''


@cache
def get_latest_chrome_major_version() -> str:
    """
    Get the latest Chrome major version.

    Returns
    -------
    str
        The latest major version number as a string.
    """
    return cast(
        'str',
        requests.get(
            ('https://versionhistory.googleapis.com/v1/chrome/platforms/win/channels/stable/'
             'versions'),
            timeout=5,
        ).json()['versions'][0]['version'].split('.')[0],
    )


@cache
def generate_chrome_user_agent(os: str = 'Windows NT 10.0; Win64; x64') -> str:
    """
    Get a Chrome user agent.

    Parameters
    ----------
    os : str
        The operating system. Default is ``'Windows NT 10.0; Win64; x64'``.

    Returns
    -------
    str
        A Chrome user agent string.
    """
    last_major = get_last_chrome_major_version() or get_latest_chrome_major_version()
    return (f'Mozilla/5.0 ({os}) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{last_major}.0.0.0'
            ' Safari/537.36')



"""Media commands."""

from __future__ import annotations

from operator import itemgetter
from pathlib import Path
from shlex import quote
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast
import json
import logging
import re
import subprocess as sp
import unicodedata

from bascom import setup_logging
from deltona.constants import CONTEXT_SETTINGS
from deltona.io import make_sfv, unpack_ebook
from deltona.media import (
    add_info_json_to_media_file,
    archive_dashcam_footage,
    cddb_query,
    create_static_text_video,
    get_info_json,
    hlg_to_sdr,
    supported_audio_input_formats,
)
from deltona.string import underscorize
from deltona.system import IS_WINDOWS, wait_for_disc
from deltona.ultraiso import InsufficientArguments, run_ultraiso
from deltona.utils import TIMES_RE, add_cdda_times
from send2trash import send2trash
from typing_extensions import override
import click

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

log = logging.getLogger(__name__)

_T = TypeVar('_T', bound=str)


class _CDDATimeStringParamType(click.ParamType):
    name = 'cdda_time_string'

    @override
    def convert(self, value: _T, param: click.Parameter | None, ctx: click.Context | None) -> _T:
        if TIMES_RE.match(value):
            return value
        self.fail(f'{value!r} is not a valid CDDA time string.', param, ctx)
        return None  # type: ignore[unreachable] # pragma: no cover


@click.command(context_settings=CONTEXT_SETTINGS,
               epilog='Example invocation: add-cdda-times 01:02:73 02:05:09')
@click.argument('times', nargs=-1, type=_CDDATimeStringParamType())
def add_cdda_times_main(times: tuple[str, ...]) -> None:
    """Add CDDA timestamps together.

    A CDDA timestamp is 3 zero-prefixed integers MM:SS:FF, separated by colons. FF is the number of
    frames out of 75.
    """  # noqa: DOC501
    if (result := add_cdda_times(times)) is None:
        raise click.Abort
    click.echo(result)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('drive_path',
                type=click.Path(exists=True, dir_okay=False, writable=True, path_type=Path))
@click.option('-w',
              '--wait-time',
              type=float,
              default=1.0,
              help='Wait time in seconds.',
              metavar='TIME')
def wait_for_disc_main(drive_path: Path, wait_time: float = 1.0) -> None:
    """Wait for a disc in a drive to be ready."""  # noqa: DOC501
    if not wait_for_disc(str(drive_path), sleep_time=wait_time):
        raise click.Abort


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--appid', metavar='STRING', help='Application ID.')
@click.option('--preparer', metavar='STRING', help='Preparer.')
@click.option('--publisher', metavar='STRING', help='Publisher.')
@click.option('--sysid', metavar='STRING', help='System ID.')
@click.option('--volset', metavar='STRING', help='Volume Set ID.', type=int)
@click.option('--volume', metavar='STRING', help='Volume label.')
@click.option(
    '--bootfile',
    metavar='FILENAME',
    help='Set boot file.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option('--bootinfotable', is_flag=True, help='Generate boot information table in boot file.')
@click.option('--optimize',
              is_flag=True,
              help='Optimise file systems by coding same files only once.')
@click.option(
    '-f',
    '--file',
    'files',
    metavar='FILENAME',
    help='Add one file or folder (include folder name and all files and folders under it).',
    multiple=True,
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    '--dir',
    'dirs',
    metavar='DIRNAME',
    multiple=True,
    help='Add all files and folders under given directory (not include directory name itself).',
    type=click.Path(exists=True, path_type=Path),
)
@click.option('--newdir', metavar='DIRNAME', help='Create a new directory.')
@click.option('-c', '--chdir', metavar='DIRNAME', help='Change current directory in ISO image.')
@click.option(
    '-r',
    '--rmdir',
    metavar='FILENAME',
    help='Remove a file or folder from ISO image (full path should be specified).',
)
@click.option(
    '--hide',
    metavar='FILENAME',
    help='Set hidden attribute of a file or folder (full path should be specified).',
)
@click.option(
    '--ahide',
    metavar='FILENAME',
    help='Set advanced hidden attribute of a file or folder (full path should be specified).',
)
@click.option(
    '-i',
    '--input',
    'input_',
    metavar='FILENAME',
    help='Input ISO image.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option('-o', '--output', metavar='FILENAME', help='Output ISO image.')
@click.option(
    '--bin2iso',
    metavar='FILENAME',
    help='Convert input CD/DVD image to ISO format.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    '--dmg2iso',
    metavar='FILENAME',
    help='Convert input DMG image to ISO format.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    '--bin2isz',
    metavar='FILENAME',
    help='Compress input CD/DVD image to ISZ format.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option('--compress', help='Set compression level.', type=click.IntRange(1, 16))
@click.option('--encrypt', help='Set encryption method.', type=click.IntRange(1, 3))
@click.option(
    '--password',
    metavar='PASSWORD',
    help='Set ISZ password.',
    prompt_required=True,
    hide_input=True,
)
@click.option('--split', metavar='SIZE', help='Set segment size in bytes.', type=int)
@click.option(
    '--list',
    'list_',
    metavar='FILENAME',
    help='Create a list of files and folders in an ISO image.',
    type=click.Path(dir_okay=False, path_type=Path),
)
@click.option(
    '--get',
    metavar='FILENAME',
    help='Set a file or folder (full path should be specified) to be extracted.',
)
@click.option('--extract', metavar='DIRNAME', help='Extract ISO image to specified directory.')
@click.option(
    '--cmd',
    metavar='FILENAME',
    help='Read arguments from a text file.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option('-l',
              '--ilong',
              is_flag=True,
              help='Long filename for ISO 9660 volume, up to 31 chars.')
@click.option('--imax', is_flag=True, help='Max filename for ISO 9660 volume, up to 207 chars.')
@click.option('--vernum', is_flag=True, help='Include file version number.')
@click.option('--lowercase', is_flag=True, help='Allow lowercase letter.')
@click.option('--joliet', is_flag=True, help='Create Joliet volume.')
@click.option('--jlong', is_flag=True, help='Long filename for joliet volume, up to 103 chars.')
@click.option('--rockridge', is_flag=True, help='Create RockRidge volume.')
@click.option('--udf', is_flag=True, help='Create UDF volume.')
@click.option('--hfs', is_flag=True, help='Create Apple HFS volume.')
@click.option(
    '--udfdvd',
    is_flag=True,
    help='Create UDF DVD image (this option will overwrite all other volume settings).',
)
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
def ultraiso_main(  # noqa: PLR0913, PLR0917
    ahide: str | None = None,
    appid: str | None = None,
    bin2iso: Path | None = None,
    bin2isz: Path | None = None,
    bootfile: Path | None = None,
    chdir: str | None = None,
    cmd: str | None = None,
    compress: int | None = None,
    dirs: Sequence[Path] | None = None,
    dmg2iso: Path | None = None,
    encrypt: int | None = None,
    extract: str | None = None,
    files: Sequence[Path] | None = None,
    get: str | None = None,
    hide: str | None = None,
    input_: Path | None = None,
    list_: Path | None = None,
    newdir: str | None = None,
    output: str | None = None,
    password: str | None = None,
    pn: int | None = None,
    prefix: str | None = None,
    preparer: str | None = None,
    publisher: str | None = None,
    rmdir: str | None = None,
    split: int | None = None,
    sysid: str | None = None,
    volset: int | None = None,
    volume: str | None = None,
    *,
    bootinfotable: bool = False,
    hfs: bool = False,
    ilong: bool = False,
    imax: bool = False,
    jlong: bool = False,
    joliet: bool = False,
    lowercase: bool = False,
    optimize: bool = False,
    rockridge: bool = False,
    udf: bool = False,
    udfdvd: bool = False,
    vernum: bool = False,
    debug: bool = False,
) -> None:
    """
    CLI interface to UltraISO.

    On non-Windows, runs UltraISO via Wine.
    """  # noqa: DOC501
    kwargs = {'prefix': prefix} if prefix and not IS_WINDOWS else {}
    setup_logging(debug=debug, loggers={'deltona': {}})
    try:
        run_ultraiso(
            add_dirs=dirs or [],
            add_files=files or [],
            bin2iso=bin2iso,
            dmg2iso=dmg2iso,
            bootfile=bootfile,
            bootinfotable=bootinfotable,
            optimize=optimize,
            cmd=cmd,
            chdir=chdir,
            newdir=newdir,
            rmdir=rmdir,
            hfs=hfs,
            jlong=jlong,
            joliet=joliet,
            rockridge=rockridge,
            udf=udf,
            udfdvd=udfdvd,
            ahide=ahide,
            hide=hide,
            pn=cast('Any', pn),
            appid=appid,
            preparer=preparer,
            publisher=publisher,
            sysid=sysid,
            volset=volset,
            volume=volume,
            input=input_,
            bin2isz=bin2isz,
            compress=cast('Any', compress),
            encrypt=cast('Any', encrypt),
            password=password,
            split=split,
            output=output,
            extract=extract,
            get=get,
            list_=list_,
            ilong=ilong,
            imax=imax,
            lowercase=lowercase,
            vernum=vernum,
            **kwargs,
        )
    except (InsufficientArguments, sp.CalledProcessError) as e:
        raise click.Abort from e
    except FileNotFoundError as e:
        click.echo('Is UltraISO installed?')
        raise click.Abort from e


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.argument('device')
def supported_audio_input_formats_main(device: str, *, debug: bool = False) -> None:
    """Get supported input formats and sample rates by invoking ffmpeg."""  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}})
    try:
        for format_, rate in supported_audio_input_formats(device):
            click.echo(f'{format_} @ {rate}')
    except OSError as e:
        click.echo('Likely invalid device name.', err=True)
        raise click.Abort from e


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1)
def add_info_json_main(filename: Sequence[Path], *, debug: bool = False) -> None:
    """Embed info.json in a media file."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    for f in filename:
        add_info_json_to_media_file(f, debug=debug)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
def display_info_json_main(filename: Path, *, debug: bool = False) -> None:
    """Display embedded info.json in a media file."""  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}})
    try:
        click.echo(get_info_json(filename, raw=True))
    except NotImplementedError as e:
        raise click.Abort from e
    except sp.CalledProcessError as e:
        click.echo(e.stdout)
        click.echo(e.stderr)
        raise click.Abort from e


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-V', '--videotoolbox', is_flag=True, help='Use VideoToolbox.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-f', '--font', default='Roboto', help='Font to use.')
@click.option('-n', '--nvenc', is_flag=True, help='Use NVENC.')
@click.option(
    '-o',
    '--output',
    'output_file',
    type=click.Path(dir_okay=False, path_type=Path),
    help='Output file.',
)
@click.option('-s', '--font-size', type=int, default=150, help='Font size in pt.')
@click.argument('audio_filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument('text', nargs=-1)
def audio2vid_main(
    audio_filename: Path,
    text: str,
    font: str = 'Roboto',
    font_size: int = 150,
    output_file: Path | None = None,
    *,
    debug: bool = False,
    nvenc: bool = False,
    videotoolbox: bool = False,
) -> None:
    """Create a video with static text and audio."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    create_static_text_video(
        audio_filename,
        ' '.join(text),
        font,
        font_size,
        output_file,
        debug=debug,
        nvenc=nvenc,
        videotoolbox=videotoolbox,
    )


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filenames', nargs=-1)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def mvid_rename_main(filenames: tuple[str, ...], *, debug: bool = False) -> None:
    """Rename music video files."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    for filename in filenames:
        path = Path(filename).resolve(strict=True)
        if not path.is_dir():
            log.debug('Ignored: %s', path)
            continue
        try:
            src = path / f'{path.name.lower()}.mkv'
            target = (path / f'../{path.name}.mkv').resolve()
            log.debug('%s -> %s', src, target)
            src.rename(target)
            send2trash(path)
        except Exception:
            log.exception('Exception with file %s.', path)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('args', nargs=-1)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-H', '--host', help='CDDB hostname.', metavar='HOST')
def cddb_query_main(args: tuple[str, ...], host: str | None = None, *, debug: bool = False) -> None:
    """
    Display a CDDB result in a simple JSON format.

    Does not handle if result is not an exact match.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    click.echo(json.dumps(cddb_query(' '.join(args), host=host)._asdict(), indent=2,
                          sort_keys=True))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('paths', type=click.Path(exists=True, file_okay=False, path_type=Path), nargs=-1)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-D', '--delete-paths', help='Delete paths after extraction.', is_flag=True)
def ke_ebook_ex_main(paths: Sequence[Path],
                     *,
                     debug: bool = False,
                     delete_paths: bool = False) -> None:
    """Extract ebooks from RARs within Zip files."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    for path in paths:
        unpack_ebook(path)
    if delete_paths:
        for path in paths:
            send2trash(path)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('front_dir', type=click.Path(exists=True, dir_okay=True, path_type=Path))
@click.argument('rear_dir', type=click.Path(exists=True, dir_okay=True, path_type=Path))
@click.argument('output_dir', type=click.Path(dir_okay=True, path_type=Path), default=Path())
@click.option('--clip-length', help='Clip length in minutes.', type=int, default=3)
@click.option('--crf', type=int, default=26, help='Constant rate factor.')
@click.option('--hwaccel', help='-hwaccel string for ffmpeg.', default='auto')
@click.option('--level', help='Level (HEVC).', default='auto')
@click.option('--no-fix-groups', help='Disable group discrepancy resolution.', is_flag=True)
@click.option('--no-hwaccel', help='Disable hardware decoding.', is_flag=True)
@click.option('--no-rear-crop', is_flag=True, help='Disable rear video cropping.')
@click.option('--no-setpts', is_flag=True, help='Disable use of setpts.')
@click.option('--preset', help='Output preset (various codecs).', default='slow')
@click.option('--rear-crop', default='1920:1020:0:0', help='Crop string for the rear camera view.')
@click.option('--rear-view-scale-divisor',
              default=2.5,
              type=float,
              help='Scaling divisor for rear view.')
@click.option('--setpts',
              help='setpts= string. Defaults to speeding video by 4x.',
              default='0.25*PTS')
@click.option('--tier', help='Tier (HEVC).', default='high')
@click.option(
    '--time-format',
    metavar='FORMAT',
    help='Time format to parse from video files.',
    default='%Y%m%d%H%M%S',
)
@click.option('--video-bitrate', default='0k', help='Video bitrate.', metavar='BITRATE')
@click.option(
    '--video-decoder',
    default='hevc_cuvid',
    help='Video decoder (for hardware decoding only).',
    metavar='DECODER',
)
@click.option('--video-encoder', default='libx265', help='Video encoder.', metavar='ENCODER')
@click.option('--video-max-bitrate',
              default='30M',
              help='Maximum video bitrate.',
              metavar='BITRATE')
@click.option('-D', '--no-delete', is_flag=True, help='Do not delete original files.')
@click.option(
    '-M',
    '--match-regexp',
    help='Regular expression to find the date string.',
    default=r'^(\d+)_.*',
    metavar='RE',
)
@click.option('-O', '--overwrite', is_flag=True, help='Overwrite existing files.')
@click.option('-T', '--temp-dir', help='Temporary directory for processing.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def encode_dashcam_main(  # noqa: PLR0913, PLR0917
    front_dir: Path,
    rear_dir: Path,
    output_dir: Path,
    clip_length: int = 3,
    crf: int = 26,
    hwaccel: str = 'auto',
    level: str = 'auto',
    match_regexp: str = r'^(\d+)_.*',
    preset: str = 'p7',
    rear_crop: str = '1920:1020:0:0',
    rear_view_scale_divisor: float = 2.5,
    setpts: str = '0.25*PTS',
    temp_dir: str | None = None,
    tier: str = 'high',
    time_format: str = '%Y%m%d%H%M%S',
    video_bitrate: str = '0k',
    video_decoder: str = 'hevc_cuvid',
    video_encoder: str = 'hevc_nvenc',
    video_max_bitrate: str = '15M',
    *,
    debug: bool = False,
    no_delete: bool = False,
    no_fix_groups: bool = False,
    no_hwaccel: bool = False,
    no_rear_crop: bool = False,
    no_setpts: bool = False,
    overwrite: bool = False,
) -> None:
    """
    Batch encode dashcam footage, merging rear and front camera footage.

    This command's defaults are intended for use with Red Tiger dashcam output and file structure.

    The rear camera view will be placed in the bottom right of the video scaled by dividing the
    width and height by the --rear-view-scale-divisor value specified. It will also be cropped using
    the --rear-crop value unless --no-rear-crop is passed.

    Files are automatically grouped using the regular expression passed with -M/--match-regexp. This
    RE must contain at least one group and only the first group will be considered. Make dubious use
    of non-capturing groups if necessary. The captured group string is expected to be usable with
    the time format specified with --time-format (see strptime documentation at
    https://docs.python.org/3/library/datetime.html#datetime.datetime.strptime).

    In many cases, the camera leaves behind stray rear camera files (usually no more than one per
    group and always a video without a matching front video file the end). These are automatically
    ignored if possible. This behaviour can be disabled by passing --no-fix-groups.

    Original files whose content is successfully converted are sent to the wastebin.

    Example use:

        encode-dashcam Movie_F/ Movie_R/ ~/output_dir
    """  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}})
    if Path(front_dir).resolve(strict=True) == Path(rear_dir).resolve(strict=True):
        click.echo('Front and rear directories are the same.', err=True)
        raise click.Abort
    archive_dashcam_footage(
        front_dir,
        rear_dir,
        output_dir,
        allow_group_discrepancy_resolution=not no_fix_groups,
        clip_length=clip_length,
        crf=crf,
        hwaccel=None if no_hwaccel else hwaccel,
        level=level,
        match_re=match_regexp,
        no_delete=no_delete,
        overwrite=overwrite,
        preset=preset,
        rear_crop=None if no_rear_crop else rear_crop,
        rear_view_scale_divisor=rear_view_scale_divisor,
        setpts=None if no_setpts else setpts,
        temp_dir=temp_dir,
        tier=tier,
        time_format=time_format,
        video_bitrate=video_bitrate,
        video_decoder=video_decoder,
        video_encoder=video_encoder,
        video_max_bitrate=video_max_bitrate,
    )


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument('output', type=click.Path(dir_okay=False, path_type=Path), required=False)
@click.option('--codec',
              help='Video codec.',
              type=click.Choice(('libx264', 'libx265')),
              default='libx265')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('--crf', help='CRF value.', type=int, default=20)
@click.option('--delete-after', help='Send processed file to wastebin.', is_flag=True)
@click.option('-f', '--fast', help='Use fewer filters (lower quality).', is_flag=True)
def hlg2sdr_main(
    filename: Path,
    output: Path | None,
    crf: int = 20,
    codec: Literal['libx264', 'libx265'] = 'libx265',
    *,
    debug: bool = False,
    delete_after: bool = False,
    fast: bool = False,
) -> None:
    """Convert a HLG video to SDR."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    hlg_to_sdr(filename, crf, codec, output, fast=fast, delete_after=delete_after)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option(
    '--input-json',
    help='Input JSON file.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
def tbc2srt_main(filename: Path, input_json: Path | None = None, *, debug: bool = False) -> None:
    """
    Convert VBI data in a ld-decode/vhs-decode TBC file to SubRip format.

    Requires the following: ld-process-vbi, ld-export-metadata, scc2raw.pl, ccextractor.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    p_filename = Path(filename)
    scc_file = p_filename.parent / f'{p_filename.stem}.scc'
    bin_file = p_filename.parent / f'{p_filename.stem}.bin'
    output_json_file = p_filename.parent / f'{p_filename.stem}.json'
    input_json = input_json or p_filename.parent / 'input.json'
    cmd: tuple[str, ...] = (
        'ld-process-vbi',
        '--input-json',
        str(input_json),
        '--output-json',
        str(output_json_file),
        str(filename),
    )
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    cmd = ('ld-export-metadata', '--closed-captions', str(scc_file), str(output_json_file))
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    cmd = ('scc2raw.pl', str(scc_file))
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    cmd = ('ccextractor', '-in=raw', str(bin_file))
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    send2trash([scc_file, bin_file, output_json_file, input_json])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('directory', type=click.Path(exists=True, file_okay=False, path_type=Path))
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def flac_dir_finalize_main(directory: Path, *, debug: bool = False) -> None:
    """Finalise a FLAC album directory."""
    def get_flac_tags(flac_path: Path) -> Iterator[tuple[str, str]]:
        def _split_eq(x: str) -> tuple[str, str] | None:
            y = x.split('=', 1)
            if len(y) == 2:  # noqa: PLR2004
                return y[0].lower(), y[1]
            return None

        return (y for y in (_split_eq(x) for x in sp.run(
            ('metaflac', '--export-tags-to=-', str(flac_path)),
            stdout=sp.PIPE,
            text=True,
            check=False,
        ).stdout.splitlines()) if y is not None)

    def remove_accents(s: str) -> str:
        return ''.join(c for c in unicodedata.normalize('NFD', re.sub(r'[Øø]', 'o', s))
                       if unicodedata.category(c) != 'Mn')

    def sanitize_for_filename(s: str) -> str:
        return underscorize(
            re.sub(
                r'[!&"\'$;`^,#\?%=\.,°±¡¯¬«ª²³¨§¦¥¤£¢¿¾½¼»þ÷¹¸·¶µ´]',  # noqa: RUF001
                '',
                re.sub(
                    r'[/\|:\\\*\+@~]',
                    '-',
                    re.sub(
                        r'[<{\(\[]',
                        '(',
                        re.sub(
                            r'[>}\)\]]',
                            ')',
                            re.sub(r'\s&\s', ' and ', remove_accents(s), flags=re.IGNORECASE),
                        ),
                    ),
                ),
            ))

    setup_logging(debug=debug, loggers={'deltona': {}})
    path = Path(directory).resolve(strict=True)
    flac_files = ((path / x) for x in Path(directory).iterdir() if x.name.endswith('.flac'))
    new_flac_files: list[Path] = []
    imgs = ((path / x) for x in Path(directory).iterdir()
            if re.search(r'\.(?:jpe?g|png|gif|webp)', str(x)) is not None)
    misc_files_prefix = f'00-{path.name.lower()}'
    img_prefix = '-'.join(misc_files_prefix.split('-')[:-1])
    out_m3u = path / f'{misc_files_prefix}.m3u'
    out_sfv = path / f'{misc_files_prefix}.sfv'
    for flac in flac_files:
        tracknumber, artist, title = itemgetter('tracknumber', 'artist', 'title')(dict(
            get_flac_tags(flac)))
        new_fn = path / re.sub(
            r'-+',
            '-',
            (f'{int(tracknumber):02d}-{sanitize_for_filename(artist)}-'
             f'{sanitize_for_filename(title)}.flac'.lower()),
        )
        new_flac_files.append(new_fn)
        flac.rename(new_fn)
    for i, img in enumerate(sorted(imgs), start=1):
        ext = img.suffix[1:].replace('jpg', 'jpeg')
        new_fn = path / f'{img_prefix}-{i:>02}.{ext}'
        img.rename(new_fn)
    with out_m3u.open('w+', encoding='utf-8') as f:
        for flac in sorted(new_flac_files):
            f.write(f'{flac.name}\r\n')
    make_sfv(out_sfv, new_flac_files)



"""Windows/Wine-related commands."""

from __future__ import annotations

from pathlib import Path
from shlex import quote
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING
import logging
import os
import shutil
import signal
import subprocess as sp

from bascom import setup_logging
from deltona.constants import CONTEXT_SETTINGS
from deltona.string import unix_path_to_wine
from deltona.system import IS_WINDOWS, kill_wine
from deltona.ultraiso import patch_ultraiso_font
from deltona.utils import unregister_wine_file_associations
from deltona.windows import DEFAULT_DPI, Field, make_font_entry
import click
import pexpect

if TYPE_CHECKING:
    from collections.abc import Sequence
    from types import FrameType

log = logging.getLogger(__name__)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def unregister_wine_file_associations_main(*, debug: bool = False) -> None:
    """Unregister Wine file associations. Terminates all Wine processes before starting."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    unregister_wine_file_associations(debug=debug)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('prefix_name')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def wineshell_main(prefix_name: str, *, debug: bool = False) -> None:
    """
    Start a new shell with WINEPREFIX set up.

    For Bash and similar shells only.
    """  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}, 'pexpect': {}})
    target = (Path(prefix_name) if Path(prefix_name).exists() else
              Path('~/.local/share/wineprefixes').expanduser() / prefix_name)
    terminal = shutil.get_terminal_size()
    c = pexpect.spawn(os.environ.get('SHELL', '/bin/bash'), ['-i'],
                      dimensions=(terminal.lines, terminal.columns))
    c.sendline(f'export WINEPREFIX={quote(str(target))}; export PS1="{target.name}🍷$PS1"')

    def resize(sig: int, frame: FrameType | None) -> None:  # pragma: no cover  # noqa: ARG001
        terminal = shutil.get_terminal_size()
        c.setwinsize(terminal.lines, terminal.columns)

    signal.signal(signal.SIGWINCH, resize)
    c.interact(escape_character=None)
    c.close()
    if c.status is not None and c.status != 0:
        raise click.exceptions.Exit(c.status)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filepath')
def unix2wine_main(filepath: str) -> None:
    """Convert a UNIX path to an absolute Wine path."""
    click.echo(unix_path_to_wine(filepath))


@click.command(context_settings=CONTEXT_SETTINGS | {'ignore_unknown_options': True})
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option('-S',
              '--very-silent',
              help='Pass /VERYSILENT (no windows will be displayed).',
              is_flag=True)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-p', '--prefix', help='Wine prefix path or name.')
def winegoginstall_main(
    args: Sequence[str],
    filename: Path,
    prefix: str,
    *,
    debug: bool = False,
    very_silent: bool = False,
) -> None:
    """
    Silent installer for GOG InnoSetup-based releases.

    This calls the installer with the following arguments:

    .. code-block:: text

       /CLOSEAPPLICATIONS /FORCECLOSEAPPLICATIONS /NOCANCEL /NORESTART /SILENT
    """  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}})
    if 'DISPLAY' not in os.environ or 'XAUTHORITY' not in os.environ:  # pragma: no cover
        log.warning(
            'Wine will likely fail to run since DISPLAY or XAUTHORITY are not in the environment.')
    env = {
        'DISPLAY': os.environ.get('DISPLAY', ''),
        'XAUTHORITY': os.environ.get('XAUTHORITY', ''),
        'WINEDEBUG': 'fixme-all',
    }
    very_silent_args = ('/SP-', '/SUPPRESSMSGBOXES', '/VERYSILENT') if very_silent else ('/SILENT',)
    if prefix:
        env['WINEPREFIX'] = (prefix if Path(prefix).exists() else str(
            (Path('~/.local/share/wineprefixes') / prefix).expanduser()))
    cmd = (
        'wine',
        str(filename),
        '/CLOSEAPPLICATIONS',
        '/FORCECLOSEAPPLICATIONS',
        '/NOCANCEL',
        '/NORESTART',
        *very_silent_args,
        *args,
    )
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    click.echo('Be very patient especially if this release is large.', err=True)
    try:
        sp.run(cmd, check=True, env=env)
    except sp.CalledProcessError as e:
        click.echo(f'STDERR: {e.stderr}', err=True)
        click.echo(f'STDOUT: {e.stdout}', err=True)
        raise click.exceptions.Exit(e.returncode) from e


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--dpi',
              default=DEFAULT_DPI,
              type=int,
              help='DPI. This should generally be left as 96.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-f', '--font', default='Noto Sans', help='Font to use.')
@click.option('-s', '--font-size', default=9, type=int, help='Font size in points.')
def set_wine_fonts_main(dpi: int = DEFAULT_DPI,
                        font: str = 'Noto Sans',
                        font_size: int = 9,
                        *,
                        debug: bool = False) -> None:
    """
    Set all Wine fonts to be the one passed in.

    This will run on Windows but it is not recommended to try on newer than Windows 7.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    with NamedTemporaryFile(mode='w+',
                            suffix='.reg',
                            prefix='set-wine-fonts',
                            delete=False,
                            encoding='utf-8') as f:
        f.write('Windows Registry Editor Version 5.00\n\n')
        f.write(r'[HKEY_CURRENT_USER\Control Panel\Desktop\WindowMetrics]')
        f.write('\n')
        f.write(''.join(
            make_font_entry(item, font, dpi=dpi, font_size_pt=font_size) for item in Field))
        f.write('\n')
    cmd = ('wine', 'regedit', '/S', f.name) if not IS_WINDOWS else ('regedit', '/S', f.name)
    log.debug('Registry file content:\n%s', Path(f.name).read_text(encoding='utf-8').strip())
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    env = {'HOME': os.environ['HOME']}
    if 'DISPLAY' not in os.environ or 'XAUTHORITY' not in os.environ:
        log.warning(
            'UltraISO.exe will likely fail to run since DISPLAY or XAUTHORITY are not in the '
            'environment.')
    if 'WINEPREFIX' in os.environ:
        env['WINEPREFIX'] = os.environ['WINEPREFIX']
    env['DISPLAY'] = os.environ.get('DISPLAY', '')
    env['XAUTHORITY'] = os.environ.get('XAUTHORITY', '')
    env['WINEDEBUG'] = 'fixme-all'
    env['PATH'] = os.environ.get('PATH', '')
    sp.run(cmd, check=True, env=env)
    Path(f.name).unlink()
    click.echo('Fonts set. Restart Wine applications for changes to take effect.')


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option(
    '-e',
    '--exe',
    help='EXE to patch.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option('-f', '--font', default='Noto Sans Regular', help='Font to use.')
def patch_ultraiso_font_main(exe: Path | None = None,
                             font: str = 'Noto Sans',
                             *,
                             debug: bool = False) -> None:
    """Patch UltraISO's hard-coded font."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    if not exe:
        if not IS_WINDOWS:
            exe = (Path(os.environ.get('WINEPREFIX', str(Path.home() / '.wine'))) / 'drive_c' /
                   'Program Files (x86)' / 'UltraISO' / 'UltraISO.exe')
        else:
            exe = (Path(os.environ.get('PROGRAMFILES(X86)', os.environ.get('PROGRAMFILES', ''))) /
                   'UltraISO' / 'UltraISO.exe')
    patch_ultraiso_font(exe, font)


@click.command(context_settings=CONTEXT_SETTINGS)
def kill_wine_main() -> None:
    """Terminate all Wine processes."""
    kill_wine()



# ruff: noqa: FBT003, S108
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
import subprocess as sp

from deltona.ultraiso import (
    ULTRAISO_FONT_REPLACEMENT_MAX_LENGTH,
    InsufficientArguments,
    InvalidExec,
    get_ultraiso_path,
    patch_ultraiso_font,
    run_ultraiso,
    unix_path_to_wine,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_patch_ultraiso_font_success(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    backup_path = tmp_path / 'UltraISO.exebak'
    original_data = b'headerMS Sans Serif\x00footer'
    exe_path.write_bytes(original_data)
    mocker.patch(
        'deltona.ultraiso.copyfile',
        side_effect=lambda src, dst: Path(dst).write_bytes(Path(src).read_bytes()),
    )
    patch_ultraiso_font(exe_path, font_name='Noto Sans')
    assert backup_path.exists()
    patched = exe_path.read_bytes()
    assert b'Noto Sans\x00' in patched
    assert b'MS Sans Serif\x00' not in patched


def test_patch_ultraiso_font_backup_exists(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    backup_path = tmp_path / 'UltraISO.exebak'
    original_data = b'headerMS Sans Serif\x00footer'
    exe_path.write_bytes(b'should not be used')
    backup_path.write_bytes(original_data)
    mocker.patch('deltona.ultraiso.copyfile')
    patch_ultraiso_font(exe_path, font_name='Noto Sans')
    assert exe_path.read_bytes().startswith(b'headerNoto Sans\x00')


def test_patch_ultraiso_font_font_too_long(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_bytes(b'MS Sans Serif\x00')
    with pytest.raises(ValueError, match=r'Font name too long.') as e:
        patch_ultraiso_font(exe_path, font_name='A' * (ULTRAISO_FONT_REPLACEMENT_MAX_LENGTH + 1))
    assert 'Font name too long' in str(e.value)


def test_patch_ultraiso_font_file_not_found(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    with pytest.raises(FileNotFoundError):
        patch_ultraiso_font(exe_path, font_name='Noto Sans')


def test_patch_ultraiso_font_is_directory(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_dir = tmp_path / 'UltraISO_dir'
    exe_dir.mkdir()
    with pytest.raises(IsADirectoryError):
        patch_ultraiso_font(exe_dir, font_name='Noto Sans')


def test_patch_ultraiso_font_backup_is_directory(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_bytes(b'MS Sans Serif\x00')
    backup_dir = tmp_path / 'UltraISO.exebak'
    backup_dir.mkdir()
    with pytest.raises(IsADirectoryError):
        patch_ultraiso_font(exe_path, font_name='Noto Sans')


def test_patch_ultraiso_font_invalid_exec(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_bytes(b'no font here')
    with pytest.raises(InvalidExec):
        patch_ultraiso_font(exe_path, font_name='Noto Sans')


def test_unix_path_to_wine_windows(mocker: MockerFixture) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', True)
    assert unix_path_to_wine(r'C:\foo\bar') == r'C:\foo\bar'
    assert unix_path_to_wine('/unix/path') == '/unix/path'


def test_unix_path_to_wine_non_windows(mocker: MockerFixture) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mock_base = mocker.patch('deltona.ultraiso.base_unix_path_to_wine',
                             return_value='Z:\\unix\\path')
    result = unix_path_to_wine('/unix/path')
    mock_base.assert_called_once_with('/unix/path')
    assert result == 'Z:\\unix\\path'


def test_unix_path_to_wine_non_windows_str_path(mocker: MockerFixture) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mock_base = mocker.patch('deltona.ultraiso.base_unix_path_to_wine',
                             return_value='Z:\\unix\\path')
    result = unix_path_to_wine(Path('/unix/path'))
    mock_base.assert_called_once_with('/unix/path')
    assert result == 'Z:\\unix\\path'


def test_run_ultraiso_cmd_arg(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mock_run = mocker.patch('deltona.ultraiso.sp.run')
    mocker.patch('deltona.ultraiso.base_unix_path_to_wine', side_effect=lambda x: f'Z:\\{x}')
    run_ultraiso(cmd='/tmp/cmd.txt', prefix=tmp_path)
    args = mock_run.call_args[0][0]
    assert 'wine' in args
    assert str(exe_path) in args
    assert '-cmd' in args
    assert 'Z:\\/tmp/cmd.txt' in args


def test_run_ultraiso_minimum_args(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mock_run = mocker.patch('deltona.ultraiso.sp.run')
    mocker.patch('deltona.ultraiso.base_unix_path_to_wine', side_effect=lambda x: f'Z:\\{x}')
    run_ultraiso(input='/tmp/in.iso', output='/tmp/out.iso', prefix=tmp_path)
    args = mock_run.call_args[0][0]
    assert '-in' in args
    assert '-out' in args


def test_run_ultraiso_insufficient_args(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    with pytest.raises(InsufficientArguments):
        run_ultraiso(prefix=tmp_path)


def test_run_ultraiso_file_not_found(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=None)
    with pytest.raises(FileNotFoundError):
        run_ultraiso(cmd='/tmp/cmd.txt', prefix=tmp_path)


def test_run_ultraiso_all_options(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mock_run = mocker.patch('deltona.ultraiso.sp.run')
    mocker.patch('deltona.ultraiso.base_unix_path_to_wine', side_effect=lambda x: f'Z:\\{x}')
    run_ultraiso(
        input='/tmp/in.iso',
        output='/tmp/out.iso',
        add_dirs=['/tmp/dir1'],
        add_files=['/tmp/file1'],
        appid='appid',
        preparer='prep',
        publisher='pub',
        sysid='sys',
        volset=1,
        volume='vol',
        ilong=True,
        imax=True,
        lowercase=True,
        vernum=True,
        hfs=True,
        jlong=True,
        joliet=True,
        rockridge=True,
        udf=True,
        udfdvd=True,
        bootfile='/tmp/boot.img',
        bootinfotable=True,
        optimize=True,
        chdir='chdir',
        newdir='newdir',
        rmdir='rmdir',
        ahide='ahide',
        hide='hide',
        pn=2,
        bin2iso='/tmp/bin2iso',
        dmg2iso='/tmp/dmg2iso',
        bin2isz='/tmp/bin2isz',
        compress=3,
        encrypt=2,
        password='pw',
        split=1024,
        extract='/tmp/extract',
        get='get',
        list_='/tmp/list',
        prefix=tmp_path,
    )
    args = ' '.join(mock_run.call_args[0][0])
    assert '-in' in args
    assert '-out' in args
    assert '-file' in args
    assert '-directory' in args
    assert '-appid' in args
    assert '-preparer' in args
    assert '-publisher' in args
    assert '-sysid' in args
    assert '-volset' in args
    assert '-volume' in args
    assert '-ilong' in args
    assert '-imax' in args
    assert '-lowercase' in args
    assert '-vernum' in args
    assert '-hfs' in args
    assert '-jlong' in args
    assert '-joliet' in args
    assert '-rockridge' in args
    assert '-udf' in args
    assert '-udfdvd' in args
    assert '-bootfile' in args
    assert '-bootinfotable' in args
    assert '-optimize' in args
    assert '-chdir' in args
    assert '-newdir' in args
    assert '-rmdir' in args
    assert '-ahide' in args
    assert '-hide' in args
    assert '-pn' in args
    assert '-bin2iso' in args
    assert '-dmg2iso' in args
    assert '-bin2isz' in args
    assert '-compress' in args
    assert '-encrypt' in args
    assert '-password' in args
    assert '-split' in args
    assert '-extract' in args
    assert '-get' in args
    assert '-list' in args


def test_run_ultraiso_windows_env(mocker: MockerFixture, tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', True)
    mock_run = mocker.patch('deltona.ultraiso.sp.run')
    run_ultraiso(input='C:\\in.iso', output='C:\\out.iso', prefix=tmp_path)
    args = mock_run.call_args[0][0]
    assert 'wine' not in args
    assert str(exe_path) in args
    assert '-in' in args
    assert '-out' in args


def test_run_ultraiso_called_process_error_with_stderr(mocker: MockerFixture,
                                                       tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mocker.patch('deltona.ultraiso.base_unix_path_to_wine', side_effect=lambda x: f'Z:\\{x}')
    error = sp.CalledProcessError(
        1,
        ['wine'],
        stderr=('some error\nfixme: ignored\nwinemenubuilder.exe'
                '\nwine: using fast synchronization.\nreal error'),
    )
    mocker.patch('deltona.ultraiso.sp.run', side_effect=error)
    mock_log = mocker.patch('deltona.ultraiso.log')
    with pytest.raises(sp.CalledProcessError):
        run_ultraiso(input='/tmp/in.iso', output='/tmp/out.iso', prefix=tmp_path)
    logged_lines = [
        call.args[1] for call in mock_log.exception.call_args_list if len(call.args) > 1
    ]
    assert any('some error' in line for line in logged_lines)
    assert any('real error' in line for line in logged_lines)
    assert not any('fixme:' in line for line in logged_lines)
    assert not any('winemenubuilder.exe' in line for line in logged_lines)
    assert not any('wine: using fast synchronization.' in line for line in logged_lines)


def test_run_ultraiso_called_process_error_with_empty_stderr(mocker: MockerFixture,
                                                             tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mocker.patch('deltona.ultraiso.base_unix_path_to_wine', side_effect=lambda x: f'Z:\\{x}')
    error = sp.CalledProcessError(1, ['wine'], stderr=b'')
    mocker.patch('deltona.ultraiso.sp.run', side_effect=error)
    mock_log = mocker.patch('deltona.ultraiso.log')
    with pytest.raises(sp.CalledProcessError):
        run_ultraiso(input='/tmp/in.iso', output='/tmp/out.iso', prefix=tmp_path)
    assert not mock_log.exception.called


def test_run_ultraiso_warns_missing_display_xauthority(mocker: MockerFixture,
                                                       tmp_path: Path) -> None:
    exe_path = tmp_path / 'UltraISO.exe'
    exe_path.write_text('dummy')
    mocker.patch('deltona.ultraiso.get_ultraiso_path', return_value=exe_path)
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mock_run = mocker.patch('deltona.ultraiso.sp.run')
    mocker.patch('deltona.ultraiso.base_unix_path_to_wine', side_effect=lambda x: f'Z:\\{x}')
    mock_environ: dict[str, str] = {'HOME': str(tmp_path)}
    mocker.patch('deltona.ultraiso.os.environ', mock_environ)
    mock_log = mocker.patch('deltona.ultraiso.log')
    run_ultraiso(input='/tmp/in.iso', output='/tmp/out.iso', prefix=tmp_path)
    assert mock_log.warning.called
    warning_msg = mock_log.warning.call_args[0][0]
    assert 'DISPLAY or XAUTHORITY' in warning_msg
    assert mock_run.called


def test_get_ultraiso_path_windows_found(mocker: MockerFixture) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', True)
    mock_exists = mocker.patch('pathlib.Path.exists', side_effect=[True, False])
    result = get_ultraiso_path('C:/')
    assert result is not None
    assert 'UltraISO.exe' in str(result)
    assert mock_exists.call_count == 1


def test_get_ultraiso_path_windows_not_found(mocker: MockerFixture) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', True)
    mocker.patch('pathlib.Path.exists', return_value=False)
    result = get_ultraiso_path('C:/')
    assert result is None


def test_get_ultraiso_path_non_windows_found(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    # Create fake UltraISO.exe in Program Files
    pf = tmp_path / 'drive_c' / 'Program Files'
    pf.mkdir(parents=True)
    exe = pf / 'UltraISO' / 'UltraISO.exe'
    exe.parent.mkdir(parents=True)
    exe.write_text('dummy')

    def exists_side_effect(self: Path) -> bool:
        return str(self) == str(exe)

    mocker.patch('pathlib.Path.exists', new=exists_side_effect)
    result = get_ultraiso_path(tmp_path)
    assert result == exe


def test_get_ultraiso_path_non_windows_not_found(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    mocker.patch('pathlib.Path.exists', return_value=False)
    result = get_ultraiso_path(tmp_path)
    assert result is None


def test_get_ultraiso_path_checks_both_program_files(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('deltona.ultraiso.IS_WINDOWS', False)
    pf86 = tmp_path / 'drive_c' / 'Program Files (x86)' / 'UltraISO'
    pf86.mkdir(parents=True)
    exe = pf86 / 'UltraISO.exe'
    exe.write_text('dummy')

    def exists_side_effect(self: Path) -> bool:
        return str(self) == str(exe)

    mocker.patch('pathlib.Path.exists', new=exists_side_effect)
    result = get_ultraiso_path(tmp_path)
    assert result == exe



from __future__ import annotations

from typing import TYPE_CHECKING
import subprocess as sp

from deltona.commands.wine import (
    kill_wine_main,
    patch_ultraiso_font_main,
    set_wine_fonts_main,
    unix2wine_main,
    unregister_wine_file_associations_main,
    winegoginstall_main,
    wineshell_main,
)

if TYPE_CHECKING:
    from pathlib import Path

    from click.testing import CliRunner
    from pytest_mock import MockerFixture


def test_unregister_wine_file_associations_main(mocker: MockerFixture, runner: CliRunner) -> None:
    unregister = mocker.patch('deltona.commands.wine.unregister_wine_file_associations')
    result = runner.invoke(unregister_wine_file_associations_main, [])
    assert result.exit_code == 0
    unregister.assert_called_once()


def test_wineshell_main_success(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    shell = '/bin/bash'
    mock_spawn = mocker.patch('deltona.commands.wine.pexpect.spawn')
    mock_spawn.return_value = mocker.Mock(status=0)
    mocker.patch('deltona.commands.wine.os.environ', {'SHELL': shell})
    prefix = tmp_path / 'prefix'
    prefix.mkdir()
    result = runner.invoke(wineshell_main, [str(prefix)])
    assert result.exit_code == 0


def test_wineshell_main_fail(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    shell = '/bin/bash'
    mock_spawn = mocker.patch('deltona.commands.wine.pexpect.spawn')
    mock_spawn.return_value = mocker.Mock(status=1)
    mocker.patch('deltona.commands.wine.os.environ', {'SHELL': shell})
    prefix = tmp_path / 'prefix'
    prefix.mkdir()
    result = runner.invoke(wineshell_main, [str(prefix)])
    assert result.exit_code != 0


def test_unix2wine_main(mocker: MockerFixture, runner: CliRunner) -> None:
    unix_path_to_wine = mocker.patch('deltona.commands.wine.unix_path_to_wine',
                                     return_value=r'C:\foo\bar')
    result = runner.invoke(unix2wine_main, ['/foo/bar'])
    assert result.exit_code == 0
    unix_path_to_wine.assert_called_once_with('/foo/bar')
    assert r'C:\foo\bar' in result.output


def test_winegoginstall_main_success(mocker: MockerFixture, runner: CliRunner,
                                     tmp_path: Path) -> None:
    file = tmp_path / 'setup.exe'
    file.write_text('dummy')
    sp_run = mocker.patch('deltona.commands.wine.sp.run')
    result = runner.invoke(winegoginstall_main, [str(file)])
    assert result.exit_code == 0
    sp_run.assert_called_once()


def test_winegoginstall_main_logs_warning_about_missing_env_vars(mocker: MockerFixture,
                                                                 runner: CliRunner,
                                                                 tmp_path: Path) -> None:
    mock_log_warning = mocker.patch('deltona.commands.wine.log.warning')
    mocker.patch(
        'deltona.commands.wine.os.environ',
        {
            'HOME': str(tmp_path),
            'PATH': '',
        },
    )
    file = tmp_path / 'setup.exe'
    file.write_text('dummy')
    sp_run = mocker.patch('deltona.commands.wine.sp.run')
    result = runner.invoke(winegoginstall_main, [str(file)])
    assert result.exit_code == 0
    sp_run.assert_called_once()
    mock_log_warning.assert_called_once_with(
        'Wine will likely fail to run since DISPLAY or XAUTHORITY are not in the environment.')


def test_winegoginstall_main_with_prefix(mocker: MockerFixture, runner: CliRunner,
                                         tmp_path: Path) -> None:
    file = tmp_path / 'setup.exe'
    file.write_text('dummy')
    prefix = tmp_path / 'a-prefix'
    prefix.mkdir()
    sp_run = mocker.patch('deltona.commands.wine.sp.run')
    result = runner.invoke(winegoginstall_main, ['-p', str(prefix), str(file)])
    assert result.exit_code == 0
    sp_run.assert_called_once()
    assert sp_run.call_args_list[0].kwargs['env']['WINEPREFIX'] == str(prefix)


def test_winegoginstall_main_subprocess_error(mocker: MockerFixture, runner: CliRunner,
                                              tmp_path: Path) -> None:
    file = tmp_path / 'setup.exe'
    file.write_text('dummy')
    mocker.patch(
        'deltona.commands.wine.sp.run',
        side_effect=sp.CalledProcessError(1, 'cmd', stderr='err', output='out'),
    )
    result = runner.invoke(winegoginstall_main, [str(file)])
    assert result.exit_code != 0


def test_set_wine_fonts_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    sp_run = mocker.patch('deltona.commands.wine.sp.run')
    mocker.patch('deltona.commands.wine.make_font_entry', return_value='FontEntry')
    mocker.patch('deltona.commands.wine.Field', ['foo'])
    mocker.patch(
        'deltona.commands.wine.os.environ',
        {
            'HOME': str(tmp_path),
            'DISPLAY': 'x',
            'XAUTHORITY': 'y',
            'PATH': '',
            'WINEPREFIX': 'a-prefix',
        },
    )
    result = runner.invoke(set_wine_fonts_main,
                           ['--dpi', '120', '--font', 'Arial', '--font-size', '10'])
    assert result.exit_code == 0
    sp_run.assert_called_once()


def test_set_wine_fonts_main_logs_warning_about_missing_env_vars(mocker: MockerFixture,
                                                                 runner: CliRunner,
                                                                 tmp_path: Path) -> None:
    sp_run = mocker.patch('deltona.commands.wine.sp.run')
    mock_log_warning = mocker.patch('deltona.commands.wine.log.warning')
    mocker.patch('deltona.commands.wine.make_font_entry', return_value='FontEntry')
    mocker.patch('deltona.commands.wine.Field', ['foo'])
    mocker.patch(
        'deltona.commands.wine.os.environ',
        {
            'HOME': str(tmp_path),
            'PATH': '',
        },
    )
    result = runner.invoke(set_wine_fonts_main,
                           ['--dpi', '120', '--font', 'Arial', '--font-size', '10'])
    assert result.exit_code == 0
    sp_run.assert_called_once()
    mock_log_warning.assert_called_once_with(
        'UltraISO.exe will likely fail to run since DISPLAY or XAUTHORITY are not in the '
        'environment.')


def test_patch_ultraiso_font_main_with_exe(mocker: MockerFixture, runner: CliRunner,
                                           tmp_path: Path) -> None:
    mocker.patch('deltona.commands.wine.IS_WINDOWS', False)  # noqa: FBT003
    exe = tmp_path / 'UltraISO.exe'
    exe.write_text('dummy')
    patch_ultraiso_font = mocker.patch('deltona.commands.wine.patch_ultraiso_font')
    result = runner.invoke(patch_ultraiso_font_main, ['--exe', str(exe), '--font', 'Arial'])
    assert result.exit_code == 0
    patch_ultraiso_font.assert_called_once_with(exe, 'Arial')


def test_patch_ultraiso_font_main_windows_no_exe(mocker: MockerFixture, runner: CliRunner,
                                                 tmp_path: Path) -> None:
    mocker.patch('deltona.commands.wine.IS_WINDOWS', True)  # noqa: FBT003
    patch_ultraiso_font = mocker.patch('deltona.commands.wine.patch_ultraiso_font')
    result = runner.invoke(patch_ultraiso_font_main, ['--font', 'Arial'])
    assert result.exit_code == 0
    patch_ultraiso_font.assert_called_once_with(mocker.ANY, 'Arial')


def test_patch_ultraiso_font_main_without_exe(mocker: MockerFixture, runner: CliRunner,
                                              tmp_path: Path) -> None:
    patch_ultraiso_font = mocker.patch('deltona.commands.wine.patch_ultraiso_font')
    mocker.patch('deltona.commands.wine.IS_WINDOWS', False)  # noqa: FBT003
    mocker.patch('os.environ', {'WINEPREFIX': str(tmp_path)})
    result = runner.invoke(patch_ultraiso_font_main, ['--font', 'Arial'])
    assert result.exit_code == 0
    assert patch_ultraiso_font.called


def test_kill_wine_main(mocker: MockerFixture, runner: CliRunner) -> None:
    kill_wine = mocker.patch('deltona.commands.wine.kill_wine')
    result = runner.invoke(kill_wine_main, [])
    assert result.exit_code == 0
    kill_wine.assert_called_once()



from __future__ import annotations

from typing import TYPE_CHECKING

from deltona.gentoo import InvalidActiveKernelSourcePath, clean_old_kernels_and_modules
import pytest

if TYPE_CHECKING:
    from unittest.mock import Mock

    from pytest_mock import MockerFixture


@pytest.fixture
def mock_path(mocker: MockerFixture) -> Mock:
    return mocker.patch('deltona.gentoo.Path')


@pytest.fixture
def mock_rmtree(mocker: MockerFixture) -> Mock:
    return mocker.patch('deltona.gentoo.rmtree')


def test_clean_old_kernels_and_modules_success(mock_path: Mock, mock_rmtree: Mock,
                                               mocker: MockerFixture) -> None:
    mock_loc = mock_path.return_value.__truediv__.return_value
    mock_loc.exists.return_value = True
    mock_loc.is_symlink.return_value = True
    mock_loc.readlink.return_value = 'linux-5.10.0'
    mock_module_dir = mocker.MagicMock()
    mock_module_dir.is_dir.return_value = True
    mock_module_dir.__str__.return_value = 'module-1'
    mock_kernel_dir = mocker.MagicMock()
    mock_kernel_dir.is_dir.return_value = True
    mock_kernel_dir.__str__.return_value = 'linux-5.9.0'
    other_non_dir = mocker.MagicMock()
    other_non_dir.is_dir.return_value = False
    mock_path.return_value.glob.side_effect = [
        [mock_module_dir, other_non_dir],
        [mock_kernel_dir, other_non_dir],
    ]
    result = list(clean_old_kernels_and_modules())
    assert result == ['module-1', 'linux-5.9.0']
    mock_rmtree.assert_any_call(mock_module_dir)
    mock_rmtree.assert_any_call(mock_kernel_dir)


def test_clean_old_kernels_and_modules_no_symlink(mock_path: Mock) -> None:
    mock_loc = mock_path.return_value.__truediv__.return_value
    mock_loc.exists.return_value = True
    mock_loc.is_symlink.return_value = False
    with pytest.raises(InvalidActiveKernelSourcePath):
        list(clean_old_kernels_and_modules())


def test_clean_old_kernels_and_modules_file_not_found(mock_path: Mock) -> None:
    mock_loc = mock_path.return_value.__truediv__.return_value
    mock_loc.exists.return_value = False
    with pytest.raises(FileNotFoundError):
        list(clean_old_kernels_and_modules())



"""System administration commands."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, TextIO
import json
import re
import sys

from bascom import setup_logging
from deltona.constants import CONTEXT_SETTINGS
from deltona.gentoo import (
    DEFAULT_ACTIVE_KERNEL_NAME,
    DEFAULT_KERNEL_LOCATION,
    DEFAULT_MODULES_PATH,
    clean_old_kernels_and_modules,
)
from deltona.system import (
    MultipleKeySlots,
    get_kconfig_dict,
    get_kwriteconfig_commands,
    patch_macos_bundle_info_plist,
    reset_tpm_enrollment,
    slug_rename,
)
from deltona.utils import secure_move_path
from deltona.www import generate_html_dir_tree
import click

if TYPE_CHECKING:
    from collections.abc import Sequence

    from paramiko import SSHClient


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('uuids', nargs=-1)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-a', '--all', 'all_', is_flag=True, help='Reset all enrolments.')
@click.option('-f', '--force', is_flag=True, help='Apply the changes.')
@click.option(
    '--crypttab',
    type=click.Path(path_type=Path, dir_okay=False, exists=True),
    help='File to read from when passing --all.',
    default='/etc/crypttab',
)
def reset_tpm_enrollments_main(
    uuids: Sequence[str],
    crypttab: Path,
    *,
    all_: bool = False,
    debug: bool = False,
    force: bool = False,
) -> None:
    """
    Reset TPM enrolments that were created by systemd-cryptenroll -tpm2-device=auto.

    Requires root privileges to work.

    Only crypttab files with UUID= entries are supported.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    if all_:
        uuids = [
            x[1][5:]
            for x in (re.split(r'\s+', line, maxsplit=4)
                      for line in (li.strip()
                                   for li in crypttab.read_text(encoding='utf-8').splitlines())
                      if not line.startswith('#')) if 'tpm2-device=auto' in x[3] and 'UUID=' in x[1]
        ]
    for uuid in uuids:
        try:
            reset_tpm_enrollment(uuid, dry_run=not force)
        except MultipleKeySlots:  # noqa: PERF203
            click.echo(f'Cannot reset TPM enrolment for {uuid}.')
            continue


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    'path',
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
    default=DEFAULT_KERNEL_LOCATION,
)
@click.option('--active-kernel-name',
              help='Kernel name like "linux".',
              default=DEFAULT_ACTIVE_KERNEL_NAME)
@click.option(
    '-m',
    '--modules-path',
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
    help='Location where modules get installed, such as "/lib/modules".',
    default=DEFAULT_MODULES_PATH,
)
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
def clean_old_kernels_and_modules_main(
    path: Path = DEFAULT_KERNEL_LOCATION,
    modules_path: Path = DEFAULT_MODULES_PATH,
    active_kernel_name: str = DEFAULT_ACTIVE_KERNEL_NAME,
    *,
    debug: bool = False,
) -> None:
    """
    Remove inactive kernels and modules.

    By default, removes old Linux sources from /usr/src.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    for item in clean_old_kernels_and_modules(path, modules_path, active_kernel_name):
        click.echo(item)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filenames', nargs=-1)
@click.option('--no-lower', is_flag=True, help='Disable lowercase.')
@click.option('-v', '--verbose', is_flag=True, help='Enable verbose output.')
def slug_rename_main(filenames: tuple[str, ...],
                     *,
                     no_lower: bool = False,
                     verbose: bool = False) -> None:
    """Rename a file to a slugified version."""
    for name in filenames:
        target = slug_rename(name, no_lower=no_lower)
        if verbose:
            click.echo(f'{name} -> {target}')


def _get_ssh_client_cls() -> type[SSHClient]:  # pragma: no cover
    from paramiko import SSHClient  # noqa: PLC0415

    return SSHClient


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filenames', type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.argument('target')
@click.option('-C', 'compress', is_flag=True, help='Enable compression.')
@click.option('-P', '--port', type=int, default=22, help='Port.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-i', '--key', 'key_filename', type=click.File('r'), help='Private key.')
@click.option('-t', '--timeout', type=float, default=2, help='Timeout in seconds.')
@click.option(
    '-p',
    'preserve',
    is_flag=True,
    help='Preserves modification times, access times, and file mode bits from the source file.',
)
@click.option('-y',
              '--dry-run',
              is_flag=True,
              help='Do not copy anything. Use with -d for testing.')
def smv_main(
    filenames: Sequence[Path],
    target: str,
    key_filename: str,
    port: int = 22,
    timeout: float = 2,
    *,
    compress: bool = False,
    debug: bool = False,
    dry_run: bool = False,
    preserve: bool = False,
) -> None:
    """
    Secure move.

    This is similar to scp but deletes the file or directory after successful copy.

    Always test with the --dry-run/-y option.
    """
    setup_logging(debug=debug, loggers={'deltona': {}, 'paramiko': {}})
    username = target.split('@', maxsplit=1)[0] if '@' in target else None
    hostname = target.split(':', maxsplit=1)[0]
    target_dir_or_filename = target.split(':')[1]
    ssh_client_cls = _get_ssh_client_cls()
    with ssh_client_cls() as client:
        client.load_system_host_keys()
        client.connect(hostname,
                       port,
                       username,
                       compress=compress,
                       key_filename=key_filename,
                       timeout=timeout)
        for filename in filenames:
            secure_move_path(client,
                             filename,
                             target_dir_or_filename,
                             dry_run=dry_run,
                             preserve_stats=preserve)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('bundle', type=click.Path(dir_okay=True, file_okay=False, path_type=Path))
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option(
    '-E',
    '--env-var',
    'env_vars',
    help='Environment variable to set.',
    multiple=True,
    type=(str, str),
)
@click.option('-r', '--retina', is_flag=True, help='For macOS apps, force Retina support.')
def patch_bundle_main(
    bundle: Path,
    env_vars: tuple[tuple[str, str], ...],
    *,
    debug: bool = False,
    retina: bool = False,
) -> None:
    """Patch a macOS/iOS/etc bundle's Info.plist file."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    data: dict[str, Any] = {}
    if env_vars:
        data['LSEnvironment'] = dict(env_vars)
    if retina:
        data['NSHighResolutionCapable'] = True
    patch_macos_bundle_info_plist(bundle, **data)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('files', type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1)
@click.option('-a', '--all', 'all_', is_flag=True, help='Find compatible files and process them.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def kconfig_to_commands_main(files: Sequence[Path],
                             *,
                             all_: bool = False,
                             debug: bool = False) -> None:
    """Generate kwriteconfig6 commands to set (Plasma) settings from your current settings."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    if all_:
        files = [*(Path.home() / '.config').glob('*rc'), Path.home() / '.config/kdeglobals']
    for file in sorted(files):
        for cmd in get_kwriteconfig_commands(file):
            click.echo(cmd)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('files', type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1)
@click.option('-a', '--all', 'all_', is_flag=True, help='Find compatible files and process them.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def kconfig_to_json_main(files: Sequence[Path], *, all_: bool = False, debug: bool = False) -> None:
    """Convert Plasma and compatible settings (INI-style) to JSON."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    if all_:
        files = [*(Path.home() / '.config').glob('*rc'), Path.home() / '.config/kdeglobals']
    for file in sorted(files):
        click.echo(
            json.dumps(get_kconfig_dict(file) | {'_file': str(file)}, indent=2, sort_keys=True))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    'path',
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
    default='.',
)
@click.option('-d', '--depth', default=2, type=int, help='Maximum depth.', metavar='DEPTH')
@click.option('-f', '--follow-symlinks', is_flag=True, help='Follow symbolic links.')
@click.option('-o', '--output-file', type=click.File('w'), default=sys.stdout, help='Output file.')
def generate_html_dir_tree_main(path: Path,
                                *,
                                output_file: TextIO,
                                depth: int = 2,
                                follow_symlinks: bool = False) -> None:
    """Generate an HTML directory listing."""
    click.echo(generate_html_dir_tree(path, follow_symlinks=follow_symlinks, depth=depth),
               output_file)



"""Media-related utility functions."""

from __future__ import annotations

from datetime import datetime
from functools import cache
from itertools import chain
from os import utime
from pathlib import Path
from shlex import quote
from shutil import copyfile
from typing import TYPE_CHECKING, Any, ClassVar, Literal, NamedTuple, cast
import contextlib
import ctypes
import getpass
import json
import logging
import operator
import os
import re
import socket
import subprocess as sp
import tempfile

from send2trash import send2trash
import keyring
import requests

from .io import context_os_open
from .system import IS_LINUX
from .typing import ProbeDict, StrPath, assert_not_none

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

__all__ = (
    'CDDBQueryResult',
    'add_info_json_to_media_file',
    'archive_dashcam_footage',
    'cddb_query',
    'create_static_text_video',
    'ffprobe',
    'get_cd_disc_id',
    'get_info_json',
    'hlg_to_sdr',
    'is_audio_input_format_supported',
    'supported_audio_input_formats',
)

log = logging.getLogger(__name__)

_DEFAULT_FORMATS = (
    'f32be',
    'f32le',
    'f64be',
    'f64le',
    's8',
    's16be',
    's16le',
    's24be',
    's24le',
    's32be',
    's32le',
    'u8',
    'u16be',
    'u16le',
    'u24be',
    'u24le',
    'u32be',
    'u32le',
)
_DEFAULT_RATES = (
    8000,
    12000,
    16000,
    22050,
    24000,
    32000,
    44100,
    48000,
    64000,
    88200,
    96000,
    128000,
    176400,
    192000,
    352800,
    384000,
)


def supported_audio_input_formats(
    input_device: str,
    *,
    formats: Sequence[str] = _DEFAULT_FORMATS,
    rates: Sequence[int] = _DEFAULT_RATES,
) -> tuple[tuple[str, int], ...]:
    """
    Get supported input formats and sample rates by invoking ``ffmpeg``.

    For possible formats, invoke ``ffmpeg``: ``ffmpeg -formats | grep PCM | cut '-d ' -f3``.

    Parameters
    ----------
    input_device : str
        Device name. Platform specific. Examples: ``'hw:Audio'``, ``'hw:NVidia'``.

    formats : Sequence[str]
        Formats to check.

    rates : Sequence[int]
        Rates in Hz to check. The default set is taken from possible frequencies for DTS format.

    Returns
    -------
    tuple[tuple[str, int], ...]
        A tuple of ``(format, rate)`` tuples.

    Raises
    ------
    OSError
        If the device is not found or busy.
    """
    ret = []
    for format_ in formats:
        for rate in rates:
            log.debug('Checking pcm_%s @ %d.', format_, rate)
            p = sp.run(
                (
                    'ffmpeg',
                    '-hide_banner',
                    '-loglevel',
                    'info',
                    '-f',
                    'alsa',
                    '-acodec',
                    f'pcm_{format_}',
                    '-ar',
                    str(rate),
                    '-i',
                    input_device,
                ),
                text=True,
                capture_output=True,
                check=False,
            )
            all_output = p.stdout.strip() + p.stderr.strip()
            if 'Device or resource busy' in all_output or 'No such device' in all_output:
                raise OSError
            log.debug('Output: %s', all_output)
            if 'cannot set sample format 0x' in all_output or f'{rate} Hz' not in all_output:
                continue
            ret.append((format_, rate))
    return tuple(ret)


def is_audio_input_format_supported(
    input_device: str,
    format: str,  # noqa: A002
    rate: int,
) -> bool:
    """
    Check if an audio format is supported by a device.

    Parameters
    ----------
    input_device : str
        Device name. Platform specific.
    format : str
        Audio format to check, such as ``'s16le'``.
    rate : int
        Sample rate in Hz.

    Returns
    -------
    bool
        ``True`` if the format and rate are supported.
    """
    return bool(supported_audio_input_formats(input_device, formats=(format,), rates=(rate,)))


def add_info_json_to_media_file(path: StrPath,
                                info_json: StrPath | None = None,
                                *,
                                debug: bool = False) -> None:
    """
    Add yt-dlp ``info.json`` file to media file at ``path``.

    On successful completion, the ``info.json`` file will be deleted.

    This function will exist until yt-dlp embeds ``info.json`` in all formats it supports where
    possible.

    This function requires the following:

    - For FLAC, MP3, and Opus: `ffmpeg <https://ffmpeg.org/>`_.
    - For MP4: `gpac <https://gpac.io/>`_.

    Parameters
    ----------
    path : StrPath
        Path to FLAC, MP3, MP4, or Opus media file.
    info_json : StrPath | None
        Path to ``info.json`` file. If not passed, ``path`` with suffix changed to ``info.json``
        is used.
    debug : bool
        If ``True``, show ffmpeg/mkvpropedit output.
    """
    path = Path(path)
    json_path = Path(info_json) if info_json else path.with_suffix('.info.json')

    def set_date() -> None:
        with json_path.open() as fp:
            data = json.load(fp)
        try:
            upload_date = data['upload_date'].strip()
        except KeyError:
            log.debug('Upload date key not found.')
            return
        if not upload_date:
            log.debug('No upload date to set.')
            return
        log.debug('Setting date to %s.', upload_date)
        seconds = datetime.strptime(upload_date, '%Y%m%d').timestamp()  # noqa: DTZ007
        utime(path, times=(seconds, seconds))

    def mkvpropedit_add_json() -> None:
        if any(
                re.match(
                    (r"^Attachment ID \d+: type 'application/json', size \d+ bytes, "
                     r"file name 'info.json'"),
                    line,
                ) for line in sp.run(('mkvmerge', '--identify', str(path)),
                                     capture_output=True,
                                     check=True,
                                     text=True).stdout.splitlines()):
            log.warning('Attachment named info.json already exists. Not modifying file.')
            return
        log.debug('Attaching info.json to MKV.')
        sp.run(
            (
                'mkvpropedit',
                str(path),
                '--attachment-name',
                'info.json',
                '--add-attachment',
                str(json_path),
            ),
            check=True,
            capture_output=not debug,
        )
        set_date()

    def flac_mp3_add_json() -> None:
        log.debug('Attaching info.json.')
        with (
                tempfile.NamedTemporaryFile(suffix=path.suffix, delete=False, dir=path.parent) as
                tf,
                tempfile.NamedTemporaryFile(suffix='.ffmetadata',
                                            encoding='utf-8',
                                            dir=path.parent,
                                            mode='w+') as ffm,
        ):
            sp.run(
                (
                    'ffmpeg',
                    '-hide_banner',
                    '-loglevel',
                    'warning',
                    '-y',
                    '-i',
                    f'file:{path}',
                    '-f',
                    'ffmetadata',
                    f'{ffm.name}',
                ),
                check=True,
                capture_output=True,
            )
            lines = Path(ffm.name).read_text(encoding='utf-8').splitlines(keepends=True)
            escaped = re.sub(r'([=;#\\\n])', r'\\\1', json_path.read_text())
            is_mp3 = path.suffix == '.mp3'
            key = r'TXXX=info_json\=' if is_mp3 else 'info_json='
            lines.insert(1, f'{key}{escaped}\n')
            with tempfile.NamedTemporaryFile(suffix='.ffmetadata',
                                             encoding='utf-8',
                                             dir=path.parent,
                                             delete=False,
                                             mode='w+') as nfw:
                nfw.writelines(lines)
            sp.run(
                (
                    'ffmpeg',
                    '-y',
                    '-i',
                    f'file:{path}',
                    '-i',
                    f'file:{nfw.name}',
                    '-map_metadata',
                    '1',
                    '-c',
                    'copy',
                    *(('-write_id3v1', '1') if is_mp3 else ()),
                    f'file:{tf.name}',
                ),
                capture_output=not debug,
                check=True,
            )
            Path(tf.name).rename(path)
            Path(nfw.name).unlink()
        set_date()

    def mp4box_add_json() -> None:
        with contextlib.suppress(sp.CalledProcessError):
            sp.run(('MP4Box', '-rem-item', '1', str(path)), capture_output=not debug, check=True)
        sp.run(('MP4Box', '-set-meta', 'mp21', str(path)), capture_output=not debug, check=True)
        info_json_path = Path('info.json')
        copyfile(json_path, info_json_path)
        log.debug('Attaching info.json to MP4.')
        sp.run(
            (
                'MP4Box',
                '-add-item',
                (f'{info_json_path}:replace:name=youtube-dl metadata:mime=application/json:'
                 'encoding=utf8'),
                str(path),
            ),
            check=True,
            capture_output=not debug,
        )
        info_json_path.unlink()
        set_date()

    if not json_path.exists():
        log.warning('JSON path not found.')
        return
    match path.suffix.lower()[1:]:
        case 'flac' | 'mp3' | 'opus':
            flac_mp3_add_json()
        case 'm4a' | 'm4b' | 'm4p' | 'm4r' | 'm4v' | 'mp4':
            mp4box_add_json()
        case 'mkv':
            mkvpropedit_add_json()
        case _:
            return
    json_path.unlink()


def ffprobe(path: StrPath) -> ProbeDict:
    """
    Run ``ffprobe`` and decode its JSON output.

    Parameters
    ----------
    path : StrPath
        Path to the media file.

    Returns
    -------
    ProbeDict
        Parsed JSON output from ``ffprobe``.
    """
    return cast(
        'ProbeDict',
        json.loads(
            sp.run(
                (
                    'ffprobe',
                    '-v',
                    'quiet',
                    '-print_format',
                    'json',
                    '-show_format',
                    '-show_streams',
                    str(path),
                ),
                check=True,
                capture_output=True,
                text=True,
            ).stdout.strip()),
    )


def get_info_json(path: StrPath, *, raw: bool = False) -> Any:
    """
    Get ``info.json`` content in ``path``.

    Parameters
    ----------
    path : StrPath
        Path to FLAC, MP3, MP4, or Opus media file.
    raw : bool
        If ``True``, do not decode.

    Returns
    -------
    Any
        The JSON data decoded. Currently not typed.
    """
    path = Path(path)
    match path.suffix.lower()[1:]:
        case 'flac':
            out = ffprobe(path)['format']['tags']['info_json']
        case 'm4a' | 'm4b' | 'm4p' | 'm4r' | 'm4v' | 'mp4':
            out = sp.run(
                ('MP4Box', '-dump-item', '1:path=/dev/stdout', str(path)),
                check=True,
                capture_output=True,
                text=True,
            ).stdout.strip()
        case 'mkv':
            out = (sp.run(
                ('mkvextract', str(path), 'attachments', '1:/dev/stdout'),
                check=True,
                capture_output=True,
                text=True,
            ).stdout.strip().splitlines()[1])
        case 'mp3':
            out = ffprobe(path)['format']['tags']['TXXX'].replace('info_json=', '', 1)
        case 'opus':
            out = ffprobe(path)['streams'][0]['tags']['info_json']
        case _:
            raise NotImplementedError
    return out if raw else json.loads(out)


def create_static_text_video(
    audio_file: StrPath,
    text: str,
    font: str = 'Roboto',
    font_size: int = 150,
    output_file: StrPath | None = None,
    *,
    debug: bool = False,
    nvenc: bool = False,
    videotoolbox: bool = False,
) -> None:
    """
    Create a video with static text overlay from an audio file.

    Parameters
    ----------
    audio_file : StrPath
        Path to the audio file.
    text : str
        Text to overlay on the video.
    font : str
        Font name to use. Default is ``'Roboto'``.
    font_size : int
        Font size in points. Default is ``150``.
    output_file : StrPath | None
        Output file path. If ``None``, defaults to ``<audio_file>-video.mkv``.
    debug : bool
        If ``True``, show ffmpeg/ImageMagick output.
    nvenc : bool
        If ``True``, use NVENC hardware encoding.
    videotoolbox : bool
        If ``True``, use VideoToolbox hardware encoding.

    Raises
    ------
    ValueError
        If both ``nvenc`` and ``videotoolbox`` are set to ``True``.
    subprocess.CalledProcessError
        If ImageMagick or FFmpeg fails.
    """
    if nvenc and videotoolbox:
        msg = 'nvenc and videotoolbox parameters are exclusive. Only one can be set to True.'
        raise ValueError(msg)
    audio_file = Path(audio_file)
    out = (Path(output_file) if output_file else
           (Path(audio_file.parent) / f'{audio_file.stem}-video.mkv'))
    with tempfile.NamedTemporaryFile(suffix='.png', delete=False, dir=Path.cwd()) as tf:
        try:
            sp.run(
                (
                    'magick',
                    '-font',
                    font,
                    '-size',
                    '1920x1080',
                    'xc:black',
                    '-fill',
                    'white',
                    '-pointsize',
                    str(font_size),
                    '-draw',
                    f"gravity Center text 0,0 '{text}'",
                    tf.name,
                ),
                capture_output=not debug,
                check=True,
            )
        except sp.CalledProcessError:
            Path(tf.name).unlink()
            raise
    args_start: tuple[str, ...] = (
        'ffmpeg',
        '-loglevel',
        'warning',
        '-hide_banner',
        '-y',
        '-loop',
        '1',
        '-i',
        tf.name,
        '-i',
        str(audio_file),
        '-shortest',
        '-acodec',
        'copy',
    )
    if nvenc:
        args_start += (
            '-vcodec',
            'h264_nvenc',
            '-profile:v',
            'high',
            '-level',
            '1',
            '-preset',
            'llhq',
            '-coder:v',
            'cabac',
            '-b:v',
            '1M',
        )
    elif videotoolbox:
        args_start += (
            '-vcodec',
            'hevc_videotoolbox',
            '-profile:v',
            'main',
            '-level',
            '1',
            '-b:v',
            '0.5M',
        )
    else:
        args_start += ('-vcodec', 'libx265', '-crf', '20', '-level', '1', '-profile:v', 'main')
    log.debug('Output: %s', out)
    sp.run(
        (*args_start, '-pix_fmt', 'yuv420p', '-b:v', '1M', '-maxrate:v', '1M', str(out)),
        capture_output=not debug,
        check=True,
    )
    Path(tf.name).unlink()


CDROMREADTOCHDR = 0x5305
CDROMREADTOCENTRY = 0x5306
CDROM_LBA = 1
CD_MSF_OFFSET = 150
CD_FRAMES = 75
CDROM_LEADOUT = 0xAA


class CDROMMSF0(ctypes.Structure):
    if TYPE_CHECKING:
        minute: int
        second: int
        frame: int
    _fields_: ClassVar[list[tuple[str, Any]]] = [
        ('minute', ctypes.c_ubyte),
        ('second', ctypes.c_ubyte),
        ('frame', ctypes.c_ubyte),
    ]


class CDROMAddress(ctypes.Union):
    if TYPE_CHECKING:
        msf: CDROMMSF0
        lba: int
    _fields_: ClassVar[list[tuple[str, Any]]] = [('msf', CDROMMSF0), ('lba', ctypes.c_int)]


class CDROMTOCEntry(ctypes.Structure):
    if TYPE_CHECKING:
        cdte_addr: CDROMAddress
        cdte_adr: int
        cdte_ctrl: int
        cdte_datamode: int
        cdte_format: int
        cdte_track: int
    _fields_: ClassVar[list[tuple[str, Any] | tuple[str, Any, int]]] = [
        ('cdte_track', ctypes.c_ubyte),
        ('cdte_adr', ctypes.c_ubyte, 4),
        ('cdte_ctrl', ctypes.c_ubyte, 4),
        ('cdte_format', ctypes.c_ubyte),
        ('cdte_addr', CDROMAddress),
        ('cdte_datamode', ctypes.c_ubyte),
    ]


class CDROMTOCHeader(ctypes.Structure):
    if TYPE_CHECKING:
        cdth_trk0: int
        cdth_trk1: int
    _fields_: ClassVar[list[tuple[str, Any]]] = [
        ('cdth_trk0', ctypes.c_ubyte),
        ('cdth_trk1', ctypes.c_ubyte),
    ]


def get_cd_disc_id(drive: StrPath) -> str:
    """
    Calculate a CDDB disc ID.

    For Linux only.

    Parameters
    ----------
    drive : StrPath
        Drive path.

    Returns
    -------
    str
        String for use with CDDB query.

    Raises
    ------
    NotImplementedError
        If not on Linux.
    OSError
        If an ioctl call fails.
    """
    if not IS_LINUX:
        raise NotImplementedError

    def cddb_sum(n: int) -> int:
        # a number like 2344 becomes 2+3+4+4 (13)
        ret = 0
        while n > 0:
            ret += n % 10
            n //= 10
        return ret

    with context_os_open(drive, os.O_RDONLY | os.O_NONBLOCK) as fd:
        libc = ctypes.CDLL('libc.so.6', use_errno=True)
        toc_header = CDROMTOCHeader()
        if libc.ioctl(fd, CDROMREADTOCHDR, ctypes.byref(toc_header)) < 0:
            raise OSError(ctypes.get_errno())
        last: int = toc_header.cdth_trk1
        toc_entries: list[CDROMTOCEntry] = []
        for i in range(last):
            buf = CDROMTOCEntry()
            buf.cdte_track = i + 1
            buf.cdte_format = CDROM_LBA
            toc_entries.append(buf)
            if libc.ioctl(fd, CDROMREADTOCENTRY, ctypes.byref(buf)):
                raise OSError(ctypes.get_errno())
        buf = CDROMTOCEntry()
        buf.cdte_track = CDROM_LEADOUT
        buf.cdte_format = CDROM_LBA
        toc_entries.append(buf)
        if libc.ioctl(fd, CDROMREADTOCENTRY, ctypes.byref(buf)) < 0:
            raise OSError(ctypes.get_errno())
    checksum = 0
    for entry in toc_entries[:-1]:
        checksum += cddb_sum((entry.cdte_addr.lba + CD_MSF_OFFSET) // CD_FRAMES)
    total_time: int = ((toc_entries[-1].cdte_addr.lba + CD_MSF_OFFSET) // CD_FRAMES) - (
        (toc_entries[0].cdte_addr.lba + CD_MSF_OFFSET) // CD_FRAMES)
    # This expression inside f'{}' causes a Yapf parsing error
    entries = ' '.join(f'{x.cdte_addr.lba + CD_MSF_OFFSET}' for x in toc_entries[:-1])
    return (f'{(checksum % 0xFF) << 24 | total_time << 8 | last:08x} {last} '
            f'{entries} '
            f'{(toc_entries[-1].cdte_addr.lba + CD_MSF_OFFSET) // CD_FRAMES}')


class CDDBQueryResult(NamedTuple):
    """CDDB query result."""

    artist: str
    """Artist name."""
    album: str
    """Album title."""
    year: int
    """Release year."""
    genre: str
    """Genre string."""
    tracks: tuple[str, ...]
    """Track titles."""


@cache
def cddb_query(
    disc_id: str,
    *,
    accept_first_match: bool = False,
    app: str = 'deltona cddb_query',
    host: str | None = None,
    timeout: float = 5,
    username: str | None = None,
    version: str = '0.0.1',
) -> CDDBQueryResult:
    """
    Run a query against a CDDB host.

    Defaults to host in Keyring under the ``gnudb`` key and current user name.

    It is advised to ``except`` typical
    `Requests exceptions <https://requests.readthedocs.io/en/latest/>`_ when calling this.

    Parameters
    ----------
    disc_id : str
        Disc ID string from :py:func:`get_cd_disc_id`.
    accept_first_match : bool
        If ``True``, accept the first match when multiple results are returned.
    app : str
        App name.
    host : str | None
        Hostname to query.
    timeout : float
        HTTP timeout.
    username : str | None
        Username for keyring and for the ``hello`` parameter to the CDDB server.
    version : str
        Application version.

    Returns
    -------
    CDDBQueryResult
        Tuple with artist, album, year, genre, and tracks.

    Raises
    ------
    ValueError
        If the server response code is not ``'200'`` or ``'210'`` (these are CDDB codes **not**
        HTTP status codes.)
    """
    username = username or getpass.getuser()
    if not username:
        raise ValueError(username)
    host = host or keyring.get_password('gnudb', username)
    if not host:
        raise ValueError(host)
    this_host = socket.gethostname()
    hello = {'hello': f'{username} {this_host} {app} {version}', 'proto': '6'}
    server = f'http://{host}/~cddb/cddb.cgi'
    r = requests.get(
        server,
        params={
            'cmd': f'cddb query {disc_id}',
            **hello
        },
        timeout=timeout,
        headers={'user-agent': hello['hello']},
    )
    r.raise_for_status()
    log.debug('Response:\n%s', r.text.strip())
    lines = r.text.splitlines()
    first_line = lines[0].split(' ', 3)
    disc_genre = disc_year = None
    if len(lines) == 1 and first_line[0] == '200':
        _, category, disc_id, artist_title = first_line
    elif first_line[0] == '210':
        if not accept_first_match:
            log.debug('Results:\n%s', '\n'.join(lines).strip())
            raise ValueError(len(lines[1:-1]))
        category, disc_id, artist_title = lines[1].split(' ', 2)
    else:
        raise ValueError(first_line[0])
    artist, disc_title = artist_title.split(' / ', 1)
    r = requests.get(
        server,
        params={
            'cmd': f'cddb read {category} {disc_id.split(" ")[0]}',
            **hello
        },
        timeout=timeout,
    )
    r.raise_for_status()
    log.debug('Response: %s', r.text)
    tracks = {}
    disc_genre = None
    disc_year = None
    log.debug('Artist: %s', artist)
    log.debug('Album: %s', disc_title)
    for line in (x.strip() for x in r.text.splitlines()[1:]
                 if x.strip() and x[0] not in {'.', '#'}):
        field_name, value = line.split('=', 1)
        match field_name:
            case 'DTITLE':
                artist, disc_title = value.split(' / ', 1)
            case 'DYEAR':
                disc_year = int(value)
            case 'DGENRE':
                disc_genre = value
            case key:
                if key.startswith('TTITLE'):
                    tracks[assert_not_none(re.match(r'^TTITLE([^=]+).*', key)).group(1)] = value
    assert disc_genre is not None
    assert disc_year is not None
    return CDDBQueryResult(
        artist,
        disc_title,
        disc_year,
        disc_genre,
        tuple(x[1] for x in sorted(tracks.items(), key=operator.itemgetter(0))),
    )


def group_files(
    items: Iterable[str],
    clip_length: int = 3,
    match_re: re.Pattern[str] | str = r'^(\d+)_.*',
    time_format: str = '%Y%m%d%H%M%S',
) -> list[list[Path]]:
    items_sorted = sorted(items)
    groups: list[list[Path]] = []
    group: list[Path] = [Path(items_sorted[0]).resolve(strict=True)]
    groups.append(group)
    for item in items_sorted[1:]:
        p = Path(item).resolve(strict=True)
        this_dt = datetime.strptime(  # noqa: DTZ007
            assert_not_none(re.match(match_re,
                                     Path(item).name)).group(1), time_format)
        last_dt = datetime.strptime(  # noqa: DTZ007
            assert_not_none(re.match(match_re,
                                     Path(group[-1]).name)).group(1), time_format)
        diff = (this_dt - last_dt).total_seconds() // 60
        log.debug('Difference for current file %s vs last file %s: %d minutes', p, group[-1], diff)
        if diff > clip_length:
            log.debug('New group started with %s.', p)
            group = [p]
            groups.append(group)
        else:
            group.append(p)
    return groups


def archive_dashcam_footage(  # noqa: PLR0913, PLR0914
    front_dir: StrPath,
    rear_dir: StrPath,
    output_dir: StrPath,
    *,
    allow_group_discrepancy_resolution: bool = True,
    clip_length: int = 3,
    container: str = 'matroska',
    crf: int | None = 28,
    extension: str = 'mkv',
    hwaccel: str | None = 'auto',
    keep_audio: bool = False,
    level: int | str | None = 'auto',
    no_delete: bool = False,
    overwrite: bool = False,
    match_re: re.Pattern[str] | str = r'^(\d+)_.*',
    preset: str | None = 'slow',
    rear_crop: str | None = '1920:1020:0:0',
    rear_view_scale_divisor: float | None = 2.5,
    setpts: str | None = '0.25*PTS',
    temp_dir: StrPath | None = None,
    tier: str | None = 'high',
    time_format: str = '%Y%m%d%H%M%S',
    video_bitrate: str | None = '0k',
    video_decoder: str | None = 'hevc_cuvid',
    video_encoder: str = 'libx265',
    video_max_bitrate: str | None = '30M',
) -> None:
    """
    Batch encode dashcam footage, merging rear and front camera footage.

    This function's defaults are intended for use with Red Tiger dashcam output and file structure.

    The rear camera view will be placed in the bottom right of the video scaled by dividing the
    width and height by the ``rear_view_scale_divisor`` value specified. It will also be cropped
    using the ``rear_crop`` value unless it is ``None``.

    Files are automatically grouped using the regular expression passed with ``match_re``. This
    RE must contain at least one group and only the first group will be considered. Make dubious use
    of non-capturing groups if necessary. The captured group string is expected to be usable with
    the time format specified with ``time_format`` (see strptime documentation at
    https://docs.python.org/3/library/datetime.html#datetime.datetime.strptime).

    In many cases, the camera leaves behind stray rear camera files (usually no more than one per
    group and always a video without a matching front video file the end). These are automatically
    ignored if possible.

    Original files whose content is successfully converted are sent to the wastebin.

    Default parameters are set to use libx265 software encoding with CUVID hardware decoding.

    Example:

    .. code::python

        archive_dashcam_footage('Movie_F', 'Movie_R', Path.home() / 'output')

    Parameters
    ----------
    front_dir : StrPath
        Directory containing front footage.
    rear_dir : StrPath
        Directory containing rear footage.
    output_dir : StrPath
        Will be created if it does not exist including parents.
    allow_group_discrepancy_resolution : bool
        Attempt to solve grouping discrepancies (count of files) automatically.
    clip_length : int
        Clip length in minutes.
    container : str
        Container to use. Must match the extension. Passed to ffmpeg's ``-f`` option.
    crf : int | None
        Constant rate factor.
    extension : str
        Output file extension. Defaults to ``mkv``.
    hwaccel : str | None
        String passed to ffmpeg's ``-hwaccel`` option. If ``None``, do not use hardware acceleration
        for decoding.
    keep_audio : bool
        Keep audio in the output video. Defaults to ``False``.
    level : int | str | None
        Level (HEVC).
    match_re : re.Pattern[str] | str
        Regular expression used for finding the timestamp in a filename. Must contain at least one
        group and only the first group is considered.
    no_delete : bool
        Do not delete original files after successful conversion.
    overwrite : bool
        Overwrite existing files.
    preset : str | None
        Preset (various codecs).
    rear_crop : str | None
        Crop string for the rear video. See `ffmpeg crop filter`_ for more information.
    rear_view_scale_divisor : float
        Scaling divisor for rear view.
    setpts : str | None
        Change the PTS. See `ffmpeg setpts filter`_ for more information. The default is to increase
        the speed of the video up by 4x.
    temp_dir : StrPath | None
        Temporary directory root.
    tier : str | None
        Tier (HEVC).
    time_format : str
        Time format string. See `strptime() Format Codes`_ for more information.
    video_bitrate : str | None
        Video bitrate string.
    video_decoder : str | None
        Video decoder.
    video_encoder : str
        Video encoder.
    video_max_bitrate : str | None
        Maximum video bitrate.

    Raises
    ------
    ValueError
        ``zip()`` is used to group pairs of file groups and the front and rear videos. Strict mode
        is used and as such length counts must always match, unless a workaround is known. If a
        workaround cannot be used, this exception will be raised from ``zip()``.
    """  # noqa: DOC501
    front_dir = Path(front_dir)
    rear_dir = Path(rear_dir)
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    # Do not sort the dicts
    input_options: list[str] = list(
        chain(*((k, *((str(v),) if not isinstance(v, bool) and v is not None else ()))
                for k, v in ({
                    '-y': overwrite,
                    '-hwaccel': hwaccel,
                }
                             | ({
                                 '-c:v': video_decoder
                             } if hwaccel else {})).items() if v)))
    crop_str = f'crop={rear_crop},' if rear_crop else ''
    setpts_str = f'setpts={setpts}' if setpts else ''
    hevc_nvenc_options = ({
        '-cq': '29',
        '-level': level,
        '-rc-lookahead': '32',
        '-spatial_aq': '1',
        '-tier': tier,
        '-tune': 'uhq',
    } if video_encoder == 'hevc_nvenc' else {})
    main_options = {'-an': not keep_audio, '-vcodec': video_encoder, '-f': container}
    libx264_options = {'-crf': str(crf)} if crf and video_encoder == 'libx264' else {}
    libx265_options = {'-crf': str(crf)} if crf and video_encoder == 'libx265' else {}
    video_bitrate_option = {'-b:v': video_bitrate} if video_bitrate else {}
    video_max_bitrate_option = {'-maxrate:v': video_max_bitrate} if video_max_bitrate else {}
    scale_filter = ((f'[0]{crop_str}'
                     f'scale=iw/{rear_view_scale_divisor}:ih/{rear_view_scale_divisor} [pip]; '
                     f'[1][pip]overlay=main_w-overlay_w:main_h-overlay_h')
                    if crop_str and rear_view_scale_divisor else '')
    filter_complex_option = ({
        '-filter_complex': f'{scale_filter},{setpts_str}'  # Trailing comma is acceptable.
    } if scale_filter or setpts_str else {})
    preset_option = {'-preset': preset} if preset else {}
    output_options = list(
        chain(*((k, *((str(v),) if not isinstance(v, bool) and v is not None else ()))
                for k, v in (main_options
                             | preset_option
                             | filter_complex_option
                             | video_bitrate_option
                             | video_max_bitrate_option
                             | hevc_nvenc_options
                             | libx265_options
                             | libx264_options).items() if v)))
    back_groups = group_files(
        (str(x) for x in Path(rear_dir).iterdir() if not x.name.startswith('.')),
        clip_length,
        match_re,
        time_format,
    )
    front_groups = group_files(
        (str(x) for x in Path(front_dir).iterdir() if not x.name.startswith('.')),
        clip_length,
        match_re,
        time_format,
    )
    back_groups_len = len(back_groups)
    front_groups_len = len(front_groups)
    log.debug('Back group count: %d', back_groups_len)
    log.debug('Front group count: %d', front_groups_len)
    if back_groups_len != front_groups_len:
        if not allow_group_discrepancy_resolution:
            raise ValueError(back_groups_len)
        log.warning('Length of front and back groups do not match. Attempting resolution.')
        back_groups = [x for x in back_groups if len(x) > 1]
        back_groups_len = len(back_groups)
        if back_groups_len != front_groups_len:
            raise ValueError(back_groups_len)
        log.info('Possibly resolved length issue by ignoring single item rear videos.')
    # Call list(zip(...)) so strictness can be checked before looping.
    for back_group, front_group in list(zip(back_groups, front_groups, strict=True)):
        with tempfile.NamedTemporaryFile('w',
                                         dir=temp_dir,
                                         encoding='utf-8',
                                         prefix='concat-',
                                         suffix='.txt') as temp_concat:
            fg_len = len(front_group)
            bg_len = len(back_group)
            log.debug('Back group length: %d', bg_len)
            log.debug('Front group length: %d', fg_len)
            for i, item in enumerate(back_group):
                log.debug(
                    'Front: %-40s Back: %s',
                    front_group[i].name if i < fg_len else 'NOTHING',
                    item.name,
                )
            if fg_len != bg_len:
                if not allow_group_discrepancy_resolution:
                    raise ValueError(bg_len)
                log.warning('List lengths of front and back videos do not match.')
                if bg_len - fg_len == 1:
                    last = back_group.pop()
                    if not no_delete:
                        send2trash(last)
                        log.debug('Sent to wastebin: %s', last)
                    log.info('Possibly resolved length issue by ignoring last rear video in set.')
                elif fg_len - bg_len == 1:
                    last = front_group.pop()
                    if not no_delete:
                        send2trash(last)
                        log.debug('Sent to wastebin: %s', last)
                    log.info('Possibly resolved length issue by ignoring last front video in set.')
                else:
                    log.error('Cannot resolve automatically.')
            to_be_merged: list[Path] = []
            send_to_waste: list[Path] = []
            for i, (back_file, front_file) in enumerate(
                    list(zip(back_group, front_group, strict=False))):
                log.debug('Back file: %s, front file: %s', back_file, front_file)
                assert back_file != front_file
                cmd = (
                    'ffmpeg',
                    '-hide_banner',
                    *input_options,
                    '-i',
                    str(back_file),
                    '-i',
                    str(front_file),
                    *output_options,
                    '-',
                )
                send_to_waste += [front_file, back_file]
                log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
                with tempfile.NamedTemporaryFile(delete=False,
                                                 dir=temp_dir,
                                                 prefix=f'{i:04d}-',
                                                 suffix=f'.{extension}') as tf:
                    try:
                        sp.run(cmd, stdout=tf, check=True, stderr=sp.PIPE)
                    except sp.CalledProcessError as e:
                        log.exception('STDERR: %s', e.stderr.decode())
                        for path in to_be_merged:
                            path.unlink()
                        raise
                    tf_fixed = Path(tf.name).resolve(strict=True)
                    to_be_merged.append(tf_fixed)
                    temp_concat.write(f"file '{tf_fixed}'\n")
            temp_concat.flush()
            full_output_path = output_dir / front_group[0].with_suffix(f'.{extension}').name
            if not overwrite:
                suffix = 1
                while full_output_path.exists():
                    offset = 5 if suffix > 1 else 0
                    full_output_path = (
                        full_output_path.parent /
                        f'{full_output_path.stem[:-offset]}-{suffix:04d}{full_output_path.suffix}')
                    suffix += 1
            cmd = (
                'ffmpeg',
                '-hide_banner',
                '-y',
                '-f',
                'concat',
                '-safe',
                '0',
                '-i',
                temp_concat.name,
                '-c',
                'copy',
                str(full_output_path),
            )
            log.debug('Concatenating with: %s', ' '.join(quote(x) for x in cmd))
            sp.run(cmd, check=True, capture_output=True)
            for path in to_be_merged:
                path.unlink()
            if not no_delete:
                for path in send_to_waste:
                    send2trash(path)
                    log.debug('Sent to wastebin: %s', path)


def hlg_to_sdr(
    input_file: StrPath,
    crf: int = 20,
    output_codec: Literal['libx265', 'libx264'] = 'libx265',
    output_file: StrPath | None = None,
    input_args: Sequence[str] | None = None,
    output_args: Sequence[str] | None = None,
    *,
    delete_after: bool = False,
    fast: bool = False,
) -> None:
    """
    Convert an HLG HDR video to SDR using tone mapping.

    Parameters
    ----------
    input_file : StrPath
        Path to the input HLG video file.
    crf : int
        Constant rate factor. Default is ``20``.
    output_codec : Literal['libx265', 'libx264']
        Output video codec. Default is ``'libx265'``.
    output_file : StrPath | None
        Output file path. If ``None``, defaults to ``<input_file>-sdr.<ext>``.
    input_args : Sequence[str] | None
        Additional input arguments for FFmpeg.
    output_args : Sequence[str] | None
        Additional output arguments for FFmpeg.
    delete_after : bool
        If ``True``, send the input file to the wastebin after conversion.
    fast : bool
        If ``True``, use fewer filters for faster but lower quality conversion.
    """
    input_file = Path(input_file)
    vf = ((
        'zscale=t=linear:npl=100,'
        'format=gbrpf32le,'
        'zscale=p=bt709,'
        'tonemap=tonemap=hable:desat=0,'
        'zscale=t=bt709:m=bt709:r=tv,'
        'format=yuv420p'
    ) if fast else (
        'zscale=tin=arib-std-b67:min=bt2020nc:pin=bt2020:rin=tv:t=arib-std-b67:m=bt2020nc:p=bt2020:'
        'r=tv,'
        'zscale=t=linear:npl=100,'
        'format=gbrpf32le,'
        'zscale=p=bt709,'
        'tonemap=tonemap=hable:desat=0,'
        'zscale=t=bt709:m=bt709:r=tv,'
        'format=yuv420p'))
    output_file = (str(output_file) if output_file else str(
        input_file.parent / f'{input_file.stem}-sdr{input_file.suffix}'))
    cmd = (
        'ffmpeg',
        '-hide_banner',
        '-y',
        *(input_args or []),
        '-i',
        str(input_file),
        *(output_args or []),
        '-c:v',
        output_codec,
        '-crf',
        str(crf),
        '-vf',
        vf,
        '-acodec',
        'copy',
        '-movflags',
        '+faststart',
        str(output_file) if output_file else f'{input_file.stem}-sdr{input_file.suffix}',
    )
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True, capture_output=True)
    if delete_after:
        send2trash(input_file)
        log.debug('Sent to wastebin: %s', input_file)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from deltona.git import (
    convert_git_ssh_url_to_https,
    get_github_default_branch,
    merge_dependabot_pull_requests,
)
import github
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_convert_git_ssh_url_to_https() -> None:
    assert (convert_git_ssh_url_to_https('git@github.com:user/repo.git') ==
            'https://github.com/user/repo')
    assert (convert_git_ssh_url_to_https('ssh://git@github.com:user/repo.git') ==
            'https://github.com/user/repo')
    assert (convert_git_ssh_url_to_https('https://github.com/user/repo.git') ==
            'https://github.com/user/repo')


def test_get_github_default_branch(mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_repo = mocker.Mock()
    mock_github.get_repo.return_value = mock_repo
    monkeypatch.setattr('github.Github', mock_github)
    mock_repo.remote.return_value.url = 'git@github.com:user/repo.git'
    mock_github.return_value.get_repo.return_value.default_branch = 'main'
    result = get_github_default_branch(repo=mock_repo, token='fake_token')
    assert result == 'main'
    mock_github.return_value.get_repo.assert_called_once_with('user/repo')


def test_merge_dependabot_pull_requests_success(mocker: MockerFixture,
                                                monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    mock_github_repo.archived = False
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'enabled'
    mock_github_repo.get_pulls.return_value = [
        mocker.Mock(user=mocker.Mock(login='dependabot[bot]'), number=1)
    ]
    mock_github_repo.get_pull.return_value.merge.return_value.merged = True
    monkeypatch.setattr('github.Github', mock_github)
    merge_dependabot_pull_requests(token='fake_token')
    mock_github_repo.get_pull.assert_called_once_with(1)
    mock_github_repo.get_pull.return_value.merge.assert_called_once_with(merge_method='rebase')


def test_merge_dependabot_pull_requests_success_get_pull_fails(
        mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    mock_github_repo.archived = False
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'enabled'
    mock_pull = mocker.MagicMock(user=mocker.Mock(login='dependabot[bot]'), number=1)
    mock_github_repo.get_pull.side_effect = github.GithubException(400)
    mock_github_repo.get_pulls.return_value = [mock_pull]
    monkeypatch.setattr('github.Github', mock_github)
    with pytest.raises(RuntimeError):
        merge_dependabot_pull_requests(token='fake_token')
    mock_pull.as_issue.return_value.create_comment.assert_not_called()


def test_merge_dependabot_pull_requests_success_alt(mocker: MockerFixture,
                                                    monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    mock_github_repo.archived = False
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'not enabled'
    mock_github_repo.get_pulls.return_value = [
        mocker.Mock(user=mocker.Mock(login='dependabot[bot]'), number=1)
    ]
    mock_github_repo.get_pull.return_value.merge.return_value.merged = True
    monkeypatch.setattr('github.Github', mock_github)
    merge_dependabot_pull_requests(token='fake_token')
    mock_github_repo.get_pull.assert_called_once_with(1)
    mock_github_repo.get_pull.return_value.merge.assert_called_once_with(merge_method='rebase')


def test_merge_dependabot_pull_requests_no_dependabot(mocker: MockerFixture,
                                                      monkeypatch: pytest.MonkeyPatch) -> None:
    def raise_ghe(*args: Any) -> None:
        raise github.GithubException(400)

    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_github_repo.archived = False
    mock_github_repo.get_contents.side_effect = raise_ghe
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'disabled'
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    monkeypatch.setattr('github.Github', mock_github)
    merge_dependabot_pull_requests(token='fake_token')
    mock_github_repo.get_pulls.assert_not_called()


def test_merge_dependabot_pull_requests_should_raise(mocker: MockerFixture,
                                                     monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    mock_github_repo.archived = False
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'enabled'
    pull = mocker.Mock(user=mocker.Mock(login='dependabot[bot]'), number=1)
    pull.get_issue_comments.return_value = []
    mock_github_repo.get_pulls.return_value = [pull]
    mock_github_repo.get_pull.return_value = pull
    mock_github_repo.get_pull.return_value.merge.side_effect = github.GithubException(400)
    monkeypatch.setattr('github.Github', mock_github)
    with pytest.raises(RuntimeError):
        merge_dependabot_pull_requests(token='fake_token')
    mock_github_repo.get_pull.assert_called_once_with(1)
    mock_github_repo.get_pull.return_value.merge.assert_called_once_with(merge_method='rebase')


def test_merge_dependabot_pull_requests_adds_recreate_comment(
        mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_pull = mocker.Mock()
    mock_issue = mocker.Mock()
    mock_github_repo.archived = False
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'enabled'
    mock_pull.user.login = 'dependabot[bot]'
    mock_pull.number = 42
    merge_result = mocker.Mock()
    merge_result.merged = False
    mock_pull.merge.return_value = merge_result
    mock_pull.as_issue.return_value = mock_issue
    mock_pull.get_issue_comments.return_value = []
    mock_github_repo.get_pulls.return_value = [mock_pull]
    mock_github_repo.get_pull.return_value = mock_pull
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    monkeypatch.setattr('github.Github', mock_github)
    merge_dependabot_pull_requests(token='fake_token')
    mock_github_repo.get_pull.assert_called_once_with(42)
    mock_pull.merge.assert_called_once_with(merge_method='rebase')
    mock_pull.as_issue.assert_called_once()
    mock_issue.create_comment.assert_called_once_with('@dependabot recreate')


def test_merge_dependabot_pull_requests_does_not_add_duplicate_recreate_comment(
        mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    mock_github = mocker.Mock()
    mock_github_repo = mocker.Mock()
    mock_pull = mocker.Mock()
    mock_issue = mocker.Mock()
    mock_github_repo.archived = False
    mock_github_repo.security_and_analysis.dependabot_security_updates.status = 'enabled'
    mock_pull.user.login = 'dependabot[bot]'
    mock_pull.number = 42
    merge_result = mocker.Mock()
    merge_result.merged = False
    mock_pull.merge.return_value = merge_result
    mock_pull.as_issue.return_value = mock_issue
    mock_comment = mocker.Mock(body='@dependabot recreate')
    mock_pull.get_issue_comments.return_value = [mock_comment]
    mock_github_repo.get_pulls.return_value = [mock_pull]
    mock_github_repo.get_pull.return_value = mock_pull
    mock_github.return_value.get_user.return_value.get_repos.return_value = [mock_github_repo]
    monkeypatch.setattr('github.Github', mock_github)
    merge_dependabot_pull_requests(token='fake_token')
    mock_github_repo.get_pull.assert_called_once_with(42)
    mock_pull.merge.assert_called_once_with(merge_method='rebase')
    mock_issue.create_comment.assert_not_called()



"""Chromium-related commands."""

from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING
import json

from bascom import setup_logging
from deltona.chromium import fix_chromium_pwa_icon
from deltona.constants import CONTEXT_SETTINGS
from deltona.system import CHROME_DEFAULT_CONFIG_PATH, CHROME_DEFAULT_LOCAL_STATE_PATH, IS_WINDOWS
from deltona.utils import kill_processes_by_name
from deltona.www import check_bookmarks_html_urls, where_from
import click

if TYPE_CHECKING:
    from collections.abc import Sequence

    from deltona.typing import ChromeLocalState


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    'files',
    metavar='FILE',
    nargs=-1,
    type=click.Path(exists=True, dir_okay=False, readable=True, path_type=Path),
)
@click.option('-w', '--webpage', is_flag=True, help='Print the webpage URL (macOS only).')
def where_from_main(files: Sequence[Path], *, webpage: bool = False) -> None:
    """Display URL where a file was downloaded from."""
    has_multiple = len(files) > 1
    for arg in files:
        if has_multiple:
            click.echo(f'{arg}: ', nl=False)
        click.echo(where_from(arg, webpage=webpage))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    'local_state_path',
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    metavar='LOCAL_STATE_PATH',
    default=CHROME_DEFAULT_LOCAL_STATE_PATH,
)
@click.option(
    '-s',
    '--subprocess-name',
    default='chrome' if not IS_WINDOWS else 'chrome.exe',
    help='Chromium-based browser subprocess name such as "chrome".',
)
@click.option(
    '--sleep-time',
    default=0.5,
    type=float,
    help='Time to sleep after attempting to kill the browser processes in seconds.',
)
def chrome_bisect_flags_main(local_state_path: Path,
                             subprocess_name: str = 'chrome',
                             sleep_time: float = 0.5) -> None:
    """
    Determine which flag is causing an issue in Chrome or any Chromium-based browser.

    Only supports removing flags (setting back to default) and not setting them to 'safe' values.
    """  # noqa: DOC501
    flags_min_len = 2

    def start_test(flags: Sequence[str], local_state: ChromeLocalState) -> tuple[bool, str | None]:
        # Test apparatus. Returns True if:
        # - there are no more flags (problem flag not found)
        # - if there is only one flag left (problem flag possibly found)
        # - if the problematic flag exists within the passed in flags
        len_flags = len(flags)
        click.echo('Testing flags:')
        for flag in flags:
            click.echo(f'- {flag}')
        local_state['browser']['enabled_labs_experiments'] = flags
        with local_state_path.open('w+', encoding='utf-8') as f:
            json.dump(local_state, f, allow_nan=False)
        click.confirm('Start browser and test for the issue, then press enter', show_default=False)
        kill_processes_by_name(subprocess_name, sleep_time, force=True)
        at_fault = click.confirm('Did the problem occur?')
        return at_fault, flags[0] if at_fault and len_flags == 1 else None

    def do_test(flags: Sequence[str], local_state: ChromeLocalState) -> str | None:
        len_flags = len(flags)
        if len_flags < flags_min_len:
            return flags[0] if len_flags == 1 else None
        done, bad_flag = start_test(flags[:len_flags // 2], deepcopy(local_state))
        if done:
            return bad_flag or do_test(flags[:len_flags // 2], local_state)
        done, bad_flag = start_test(flags[len_flags // 2:], deepcopy(local_state))
        if done:
            return bad_flag or do_test(flags[len_flags // 2:], local_state)
        return None

    p_ls = local_state_path.resolve(strict=True)
    click.echo(f'Using "{local_state_path}".')
    with p_ls.open(encoding='utf-8') as f:
        local_state_data = json.load(f)
        flags = local_state_data['browser']['enabled_labs_experiments']
        len_flags = len(flags)
        if len_flags == 0:
            click.echo('Nothing to test.', err=True)
            raise click.Abort
    bad_flag = None
    try:
        click.confirm('Exit the browser and press enter', show_default=False)
        bad_flag = do_test(flags, local_state_data)
    except KeyboardInterrupt as e:  # pragma: no cover
        raise click.Abort from e
    finally:
        if bad_flag:
            local_state_data['browser']['enabled_labs_experiments'] = [
                x for x in local_state_data['browser']['enabled_labs_experiments'] if x != bad_flag
            ]
        with p_ls.open('w+', encoding='utf-8') as f:
            json.dump(local_state_data, f, sort_keys=True, indent=2, allow_nan=False)
        if not bad_flag:
            click.echo('Restored original "Local State".')
        else:
            click.echo(f'Saved "Local State" with "{bad_flag}" removed.')


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('app_id')
@click.argument('icon_src_uri')
@click.option('-M', '--monochrome', help='Copy icons to monochrome directory.', is_flag=True)
@click.option(
    '-c',
    '--config-path',
    help='Chromium browser configuration path. Defaults to Google Chrome.',
    default=CHROME_DEFAULT_CONFIG_PATH,
    type=click.Path(exists=True, path_type=Path),
)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-m', '--masked', help='Copy icons to masked directory.', is_flag=True)
@click.option('-p', '--profile', help='Profile name.', default='Default')
def fix_chromium_pwa_icon_main(
    config_path: Path,
    app_id: str,
    icon_src_uri: str,
    profile: str = 'Default',
    *,
    debug: bool = False,
    masked: bool = False,
    monochrome: bool = False,
) -> None:
    """
    Fix a Chromium PWA icon that failed to sync.

    For more information see https://issues.chromium.org/issues/40595456.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    fix_chromium_pwa_icon(config_path,
                          app_id,
                          icon_src_uri,
                          profile,
                          masked=masked,
                          monochrome=monochrome)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def check_bookmarks_html_main(filename: Path, *, debug: bool = False) -> None:
    """Check for URLs that are not valid any more (status 404) and redirections."""
    setup_logging(debug=debug, loggers={'deltona': {}, 'urllib3': {}})
    _, changed, not_found = check_bookmarks_html_urls(Path(filename).read_text(encoding='utf-8'))
    click.echo(f'{len(changed)} URLs changed.')
    click.echo(f'{len(not_found)} URLs resulted in 404 response.')
